# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import jwt.utils
import secrets
import time
import uuid
import re
from unittest import mock

import pytest

from planet_auth import ScopeNotGrantedTokenException, InvalidArgumentException
from planet_auth.auth import Auth
from planet_auth.auth_client import AuthClient, AuthClientConfig
from planet_auth.oidc.auth_clients.client_validator import OidcClientValidatorAuthClient
from planet_auth.oidc.auth_clients.client_credentials_flow import ClientCredentialsClientSecretAuthClient
from planet_auth.auth_exception import AuthException, InvalidTokenException
from planet_auth.oidc.multi_validator import OidcMultiIssuerValidator
from tests.test_planet_auth.unit.auth.util import StubOidcAuthClient, StubOidcClientConfig, FakeTokenBuilder
from tests.test_planet_auth.util import tdata_resource_file_path

# TEST_AUTH_SERVER = "https://blackhole.unittest.planet.com/oauth2"
TEST_PRIMARY_ISSUER = "test_primary_issuer"
TEST_PRIMARY_SIGNING_KEY = tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem")
TEST_PRIMARY_PUB_KEY = tdata_resource_file_path("keys/keypair1_pub_jwk.json")

TEST_SECONDARY_ISSUER = "test_secondary_issuer"
TEST_SECONDARY_SIGNING_KEY = tdata_resource_file_path("keys/keypair2_priv_nopassword.test_pem")
TEST_SECONDARY_PUB_KEY = tdata_resource_file_path("keys/keypair2_pub_jwk.json")

TEST_UNTRUSTED_ISSUER = "test_untrusted_issuer"
TEST_UNTRUSTED_SIGNING_KEY = tdata_resource_file_path("keys/keypair3_priv_nopassword.test_pem")
TEST_UNTRUSTED_PUB_KEY = tdata_resource_file_path("keys/keypair3_pub_jwk.json")

TEST_AUDIENCE = "test_audience"

TEST_TOKEN_TTL = 60


#
# Since the return values are governed by different standards,
# there is a lot more ways for remote validation to be considered
# invalid.  This may be redundant with testing of the lower level
# classes, but auth is so important to get right, this is OK by me.
#
class StubOidcAuthClient_ForceFailRemoteValidation(StubOidcAuthClient):
    def validate_access_token_remote(self, access_token: str) -> dict:
        raise AuthException("Forced test failure")


class StubOidcAuthClient_InvalidRemoteValidation1(StubOidcAuthClient):
    def validate_access_token_remote(self, access_token: str) -> dict:
        # Not a legal response. Errors are supposed to throw
        return None


class StubOidcAuthClient_InvalidRemoteValidation2(StubOidcAuthClient):
    def validate_access_token_remote(self, access_token: str) -> dict:
        # Not a legal response. Errors are supposed to throw
        return {}


class StubOidcAuthClient_InvalidRemoteValidation3(StubOidcAuthClient):
    def validate_access_token_remote(self, access_token: str) -> dict:
        # Not a legal response. Missing "active" key.  (Is this our job, or the job of the IntrospectionApiClient to enforce?)
        return {"some_key": "some_value"}


class StubOidcAuthClient_InvalidRemoteValidation4(StubOidcAuthClient):
    def validate_access_token_remote(self, access_token: str) -> dict:
        # Not a legal response. value is not a bool.
        return {"active": "true"}


class StubOidcAuthClient_InvalidRemoteValidation5(StubOidcAuthClient):
    def validate_access_token_remote(self, access_token: str) -> dict:
        # Not a legal response. value is not a bool.
        return {"active": False}


class MVUnitTestException(Exception):
    pass


class StubOidcAuthClient_InvalidRemoteValidation6(StubOidcAuthClient):
    def validate_access_token_remote(self, access_token: str) -> dict:
        raise MVUnitTestException("Unexpected exception class")


class StubOidcAuthClient_InvalidLocalValidation1(StubOidcAuthClient):
    def validate_access_token_local(
        self, access_token: str, required_audience: str = None, scopes_anyof: list = None
    ) -> dict:
        # Not a legal response. Errors are supposed to throw
        return None


class StubOidcAuthClient_InvalidLocalValidation2(StubOidcAuthClient):
    def validate_access_token_local(
        self, access_token: str, required_audience: str = None, scopes_anyof: list = None
    ) -> dict:
        return {}


primary_issuer_config_dict = {
    "auth_server": TEST_PRIMARY_ISSUER,
    "scopes": ["test_scp0", "test_scp1"],
    "audiences": [TEST_AUDIENCE],
    "stub_authority_ttl": TEST_TOKEN_TTL,
    "stub_authority_access_token_audience": TEST_AUDIENCE,
    "stub_authority_signing_key_file": TEST_PRIMARY_SIGNING_KEY,
    "stub_authority_pub_key_file": TEST_PRIMARY_PUB_KEY,
}
primary_issuer_config = StubOidcClientConfig(**primary_issuer_config_dict)

secondary_issuer_config_dict = {
    "auth_server": TEST_SECONDARY_ISSUER,
    "scopes": ["test_scp0", "test_scp2"],
    "audiences": [TEST_AUDIENCE],
    "stub_authority_ttl": TEST_TOKEN_TTL,
    "stub_authority_access_token_audience": TEST_AUDIENCE,
    "stub_authority_signing_key_file": TEST_SECONDARY_SIGNING_KEY,
    "stub_authority_pub_key_file": TEST_SECONDARY_PUB_KEY,
}
secondary_issuer_config = StubOidcClientConfig(**secondary_issuer_config_dict)

untrusted_issuer_config_dict = {
    "auth_server": TEST_UNTRUSTED_ISSUER,
    "scopes": ["test_scp0", "test_scp2"],
    "audiences": [TEST_AUDIENCE],
    "stub_authority_ttl": TEST_TOKEN_TTL,
    "stub_authority_access_token_audience": TEST_AUDIENCE,
    "stub_authority_signing_key_file": TEST_UNTRUSTED_SIGNING_KEY,
    "stub_authority_pub_key_file": TEST_UNTRUSTED_PUB_KEY,
}
untrusted_issuer_config = StubOidcClientConfig(**untrusted_issuer_config_dict)

bad_config_1_dict = {
    "auth_server": TEST_PRIMARY_ISSUER,
    "scopes": ["test_scp0", "test_scp1"],
    "audiences": None,  # Missing
    "stub_authority_ttl": TEST_TOKEN_TTL,
    "stub_authority_access_token_audience": TEST_AUDIENCE,
    "stub_authority_signing_key_file": TEST_PRIMARY_SIGNING_KEY,
    "stub_authority_pub_key_file": TEST_PRIMARY_PUB_KEY,
}
bad_config_1 = StubOidcClientConfig(**bad_config_1_dict)

primary_issuer = StubOidcAuthClient(primary_issuer_config)
secondary_issuer = StubOidcAuthClient(secondary_issuer_config)
untrusted_issuer = StubOidcAuthClient(untrusted_issuer_config)
bad_validator_1 = StubOidcAuthClient(bad_config_1)

primary_failRemote = StubOidcAuthClient_ForceFailRemoteValidation(primary_issuer_config)
primary_invalidRemote1 = StubOidcAuthClient_InvalidRemoteValidation1(primary_issuer_config)
primary_invalidRemote2 = StubOidcAuthClient_InvalidRemoteValidation2(primary_issuer_config)
primary_invalidRemote3 = StubOidcAuthClient_InvalidRemoteValidation3(primary_issuer_config)
primary_invalidRemote4 = StubOidcAuthClient_InvalidRemoteValidation4(primary_issuer_config)
primary_invalidRemote5 = StubOidcAuthClient_InvalidRemoteValidation5(primary_issuer_config)
# primary_invalidRemote6 = StubOidcAuthClient_InvalidRemoteValidation6(primary_issuer_config)
primary_invalidLocal1 = StubOidcAuthClient_InvalidLocalValidation1(primary_issuer_config)
primary_invalidLocal2 = StubOidcAuthClient_InvalidLocalValidation2(primary_issuer_config)

secondary_failRemote = StubOidcAuthClient_ForceFailRemoteValidation(secondary_issuer_config)
secondary_invalidRemote1 = StubOidcAuthClient_InvalidRemoteValidation1(secondary_issuer_config)
secondary_invalidRemote2 = StubOidcAuthClient_InvalidRemoteValidation2(secondary_issuer_config)
secondary_invalidRemote3 = StubOidcAuthClient_InvalidRemoteValidation3(secondary_issuer_config)
secondary_invalidRemote4 = StubOidcAuthClient_InvalidRemoteValidation4(secondary_issuer_config)
secondary_invalidRemote5 = StubOidcAuthClient_InvalidRemoteValidation5(secondary_issuer_config)
# secondary_invalidRemote6 = StubOidcAuthClient_InvalidRemoteValidation6(secondary_issuer_config)
secondary_invalidLocal1 = StubOidcAuthClient_InvalidLocalValidation1(secondary_issuer_config)
secondary_invalidLocal2 = StubOidcAuthClient_InvalidLocalValidation2(secondary_issuer_config)


class TestMultiValidator:
    # TODO: rather than monkey patch, let the validator take auth clients as arguments?
    #       This would also let us maybe handle other auth mechanisms beyond oauth?
    @staticmethod
    def _patch_primary(under_test: OidcMultiIssuerValidator, patched_auth_client: AuthClient):
        under_test._trusted[TEST_PRIMARY_ISSUER]._auth_client = patched_auth_client

    @staticmethod
    def _patch_secondary(under_test: OidcMultiIssuerValidator, patched_auth_client: AuthClient):
        under_test._trusted[TEST_SECONDARY_ISSUER]._auth_client = patched_auth_client

    def under_test__only_primary_validator(self, log_primary=True):
        under_test = OidcMultiIssuerValidator(
            trusted=[Auth.initialize_from_client(auth_client=primary_issuer)], log_result=log_primary
        )
        self._patch_primary(under_test, primary_issuer)
        return under_test

    def under_test__primary_and_secondary_validator(self, log_primary=True, log_secondary=True):
        under_test = OidcMultiIssuerValidator(
            trusted=[Auth.initialize_from_client(primary_issuer), Auth.initialize_from_client(secondary_issuer)],
            log_result=log_primary,
        )
        return under_test

    def under_test__bad_client_1(self):
        under_test = OidcMultiIssuerValidator(trusted=[Auth.initialize_from_client(bad_validator_1)])
        return under_test

    def username(self, test_case_name):
        return test_case_name + "_username"

    def assert_successful_validation_local(self, test_case_name, local_validation, required_issuer):
        # validation failure should throw. spot check some claims for good
        # measure when we need to assert validation success.
        assert TEST_AUDIENCE == local_validation.get("aud")
        assert test_case_name == local_validation.get("test_case")
        assert required_issuer == local_validation.get("iss")

    def assert_successful_validation_remote(self, test_case_name, remote_validation, required_issuer):
        # validation failure should throw. spot check some claims for good
        # measure when we need to assert validation success.
        test_username = self.username(test_case_name)
        assert remote_validation.get("active")
        assert TEST_AUDIENCE == remote_validation.get("aud")
        assert required_issuer == remote_validation.get("iss")
        assert test_username == remote_validation.get("sub")

    def test__primary_only__local_validation_only__success(self):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__only_primary_validator()
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False
        )

        self.assert_successful_validation_local(test_case_name, local_validation, TEST_PRIMARY_ISSUER)
        assert remote_validation is None

    def test_empty_access_token(self):
        # QE TC3, TC5
        under_test = self.under_test__only_primary_validator()
        with pytest.raises(InvalidArgumentException):
            local_validation, remote_validation = under_test.validate_access_token(token=None)
        with pytest.raises(InvalidArgumentException):
            local_validation, remote_validation = under_test.validate_access_token(token="")

    def test_malformed_token_1(self):
        # QE TC6 - just some random garbage.
        under_test = self.under_test__only_primary_validator()
        with pytest.raises(InvalidTokenException):
            under_test.validate_access_token(token=secrets.token_bytes(2048))

    def test_malformed_token_2(self):
        # QE TC6 - just some random garbage, but URL safe.
        under_test = self.under_test__only_primary_validator()
        with pytest.raises(InvalidTokenException):
            under_test.validate_access_token(token=secrets.token_urlsafe(2048))

    def test_malformed_token_3(self):
        # QE TC6 - random garbage, but has JWT three dot structure.
        under_test = self.under_test__only_primary_validator()
        fake_jwt = "{}.{}.{}".format(
            str(jwt.utils.base64url_encode(secrets.token_bytes(256)), encoding="utf-8"),
            str(jwt.utils.base64url_encode(secrets.token_bytes(2048)), encoding="utf-8"),
            str(jwt.utils.base64url_encode(secrets.token_bytes(256)), encoding="utf-8"),
        )

        with pytest.raises(InvalidTokenException):
            under_test.validate_access_token(token=fake_jwt)

    def test_malformed_token_4(self):
        # QE TC6 - random garbage, but looks more JWT like with encoded json.  Signature is garbage.
        under_test = self.under_test__only_primary_validator()

        fake_jwt = FakeTokenBuilder.fake_token(
            body={
                "fake_claim_1": "test claim value",
                "fake_claim_2": "test claim value",
                "iss": primary_issuer.token_builder.issuer,
            },
            header={
                "alg": primary_issuer.token_builder.signing_key_algorithm,
                "kid": primary_issuer.token_builder.signing_key_id,
            },
        )

        with pytest.raises(InvalidTokenException):
            under_test.validate_access_token(token=fake_jwt)

    def test_malformed_token_5_iss_liars(self):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)
        under_test = self.under_test__only_primary_validator()

        now = int(time.time())
        token_body = {
            "iss": primary_issuer.token_builder.issuer,
            "sub": test_username,
            "aud": primary_issuer.token_builder.audience,
            "iat": now,
            "exp": now + 100,
            "jti": str(uuid.uuid4()),
        }
        token_header = {
            "alg": primary_issuer.token_builder.signing_key_algorithm,
            "kid": primary_issuer.token_builder.signing_key_id,
        }

        # TC 0
        # Build a valid token first, just to make sure our test method is valid
        test_jwt = primary_issuer.token_builder.encode(body=token_body, extra_headers=token_header)
        under_test.validate_access_token(token=test_jwt)  # No throw

        # TC 1
        # Use the real signing key, but make the iss an invalid type.
        # You can make the argument this is still valid, because the signature is still
        # from a trusted issuer. But, we reject it based on bad structure.
        token_body["iss"] = [primary_issuer.token_builder.issuer, untrusted_issuer.token_builder.issuer]
        test_jwt = primary_issuer.token_builder.encode(body=token_body, extra_headers=token_header)
        with pytest.raises(
            InvalidTokenException,
            match=re.escape("Issuer claim ('iss') must be a of string type. 'list' type was detected."),
        ):
            under_test.validate_access_token(token=test_jwt)

        # TC 2
        # Liar token.  Untrusted issuer signing key claiming to be valid issuer
        token_body["iss"] = primary_issuer.token_builder.issuer
        test_jwt = untrusted_issuer.token_builder.encode(body=token_body, extra_headers=token_header)
        with pytest.raises(
            InvalidTokenException, match=re.escape("Signature verification failed (InvalidSignatureError)")
        ):
            under_test.validate_access_token(token=test_jwt)

        # TC 3
        # Double-talk liar.  Using the untrusted signing key, claiming to be ourselves and the trusted issuer.
        token_body["iss"] = [primary_issuer.token_builder.issuer, untrusted_issuer.token_builder.issuer]
        test_jwt = untrusted_issuer.token_builder.encode(body=token_body, extra_headers=token_header)
        with pytest.raises(
            InvalidTokenException,
            match=re.escape("Issuer claim ('iss') must be a of string type. 'list' type was detected."),
        ):
            under_test.validate_access_token(token=test_jwt)

    def test_missing_signature(self):
        # QE TC11 - JWT without a signature
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)
        under_test = self.under_test__only_primary_validator()

        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        token_parts = access_token.access_token().split(".")

        # test the test, make sure our mangling technique doesn't cause unexpected problems.
        # This should pass validation without any exceptions.
        mangled_token = "{}.{}.{}".format(token_parts[0], token_parts[1], token_parts[2])
        local_validation, remote_validation = under_test.validate_access_token(token=mangled_token)

        # real test 1 - with the last "."
        mangled_token = "{}.{}.".format(token_parts[0], token_parts[1])
        with pytest.raises(InvalidTokenException):
            local_validation, remote_validation = under_test.validate_access_token(token=mangled_token)

        # real test 2 - without the last "."
        mangled_token = "{}.{}".format(token_parts[0], token_parts[1])
        with pytest.raises(InvalidTokenException):
            local_validation, remote_validation = under_test.validate_access_token(token=mangled_token)

    def test_malformed_token_missing_issuer(self):
        # QE TC14
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)
        access_token = primary_issuer.login(
            username=test_username, extra_claims={"test_case": test_case_name}, remove_claims=["iss"]
        )

        under_test = self.under_test__only_primary_validator()
        with pytest.raises(InvalidTokenException):
            under_test.validate_access_token(token=access_token.access_token())

    def test__primary_only__local_validation_only__test_scope_validation(self):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__only_primary_validator()

        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False, scopes_anyof=None
        )
        self.assert_successful_validation_local(test_case_name, local_validation, TEST_PRIMARY_ISSUER)

        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False, scopes_anyof=[]
        )
        self.assert_successful_validation_local(test_case_name, local_validation, TEST_PRIMARY_ISSUER)

        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(),
            do_remote_revocation_check=False,
            scopes_anyof=["test_scp1", "some_scome_not_in_the_token"],
        )
        self.assert_successful_validation_local(test_case_name, local_validation, TEST_PRIMARY_ISSUER)

        with pytest.raises(ScopeNotGrantedTokenException):
            under_test.validate_access_token(
                access_token.access_token(), do_remote_revocation_check=False, scopes_anyof=["test_missing_scope"]
            )

    def test__primary_only__local_and_remote_validation__success(self):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__only_primary_validator()
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=True
        )

        self.assert_successful_validation_local(test_case_name, local_validation, TEST_PRIMARY_ISSUER)
        self.assert_successful_validation_remote(test_case_name, remote_validation, TEST_PRIMARY_ISSUER)

    @pytest.mark.parametrize("failing_locally_primary_issuer", [(primary_invalidLocal1), (primary_invalidLocal2)])
    def test__primary_only__fail_local_validation(self, failing_locally_primary_issuer):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__only_primary_validator()
        self._patch_primary(under_test, failing_locally_primary_issuer)

        with pytest.raises(AuthException):
            local_validation, remote_validation = under_test.validate_access_token(
                access_token.access_token(), do_remote_revocation_check=False
            )

    @pytest.mark.parametrize(
        "failing_remote_primary_issuer",
        [
            (primary_failRemote),
            (primary_invalidRemote1),
            (primary_invalidRemote2),
            (primary_invalidRemote3),
            (primary_invalidRemote4),
            (primary_invalidRemote5),
            # (primary_invalidRemote6),
        ],
    )
    def test__primary_only__fail_remote_validation(self, failing_remote_primary_issuer):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__only_primary_validator()
        self._patch_primary(under_test, failing_remote_primary_issuer)

        # Local check passes
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False
        )
        self.assert_successful_validation_local(test_case_name, local_validation, TEST_PRIMARY_ISSUER)
        assert remote_validation is None

        # Remote check fails
        with pytest.raises(AuthException):
            local_validation, remote_validation = under_test.validate_access_token(
                access_token.access_token(), do_remote_revocation_check=True
            )

    def test__secondary__local_auth_only_success(self):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = test_case_name + "_username"

        access_token = secondary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__primary_and_secondary_validator()
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False
        )

        self.assert_successful_validation_local(test_case_name, local_validation, TEST_SECONDARY_ISSUER)
        assert remote_validation is None

    def test__secondary__local_and_remote_auth_success(self):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = test_case_name + "_username"

        access_token = secondary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__primary_and_secondary_validator()
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=True
        )

        self.assert_successful_validation_local(test_case_name, local_validation, TEST_SECONDARY_ISSUER)
        self.assert_successful_validation_remote(test_case_name, remote_validation, TEST_SECONDARY_ISSUER)

    @pytest.mark.parametrize(
        "failing_remote_secondary_issuer",
        [
            (secondary_failRemote),
            (secondary_invalidRemote1),
            (secondary_invalidRemote2),
            (secondary_invalidRemote3),
            (secondary_invalidRemote4),
            (secondary_invalidRemote5),
            # (secondary_invalidRemote6),
        ],
    )
    def test__secondary__fail_remote_validation(self, failing_remote_secondary_issuer):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        access_token = secondary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__primary_and_secondary_validator()
        self._patch_secondary(under_test, failing_remote_secondary_issuer)

        # Local check passes
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False
        )
        self.assert_successful_validation_local(test_case_name, local_validation, TEST_SECONDARY_ISSUER)
        assert remote_validation is None

        # Remote check fails
        with pytest.raises(AuthException):
            local_validation, remote_validation = under_test.validate_access_token(
                access_token.access_token(), do_remote_revocation_check=True
            )

    def test_check_auth_client_types(self):
        #  test that we check the types of passed auth clients to the ctor
        with pytest.raises(AuthException):
            OidcMultiIssuerValidator(
                trusted=[
                    Auth.initialize_from_config(
                        client_config=AuthClientConfig.from_file(
                            tdata_resource_file_path("auth_client_configs/utest/planet_legacy.json")
                        )
                    )
                ]
            )
        with pytest.raises(AuthException):
            OidcMultiIssuerValidator(
                trusted=[
                    Auth.initialize_from_client(auth_client=primary_issuer),
                    Auth.initialize_from_config(
                        client_config=AuthClientConfig.from_file(
                            tdata_resource_file_path("auth_client_configs/utest/static_api_key.json")
                        )
                    ),
                ],
            )

    def test_no_repeat_issuers(self):
        with pytest.raises(AuthException):
            OidcMultiIssuerValidator(
                trusted=[Auth.initialize_from_client(primary_issuer), Auth.initialize_from_client(primary_issuer)],
            )
        with pytest.raises(AuthException):
            OidcMultiIssuerValidator(
                trusted=[
                    Auth.initialize_from_client(primary_issuer),
                    Auth.initialize_from_client(secondary_issuer),
                    Auth.initialize_from_client(primary_issuer),
                ],
            )

    def test_ignore_falsy_issuers_in_direct_construction(self):
        under_test = OidcMultiIssuerValidator(
            trusted=["", Auth.initialize_from_client(primary_issuer), None],
        )
        assert len(under_test._trusted) == 1

        under_test = OidcMultiIssuerValidator(
            trusted=[
                "",
                Auth.initialize_from_client(primary_issuer),
                Auth.initialize_from_client(secondary_issuer),
                None,
            ],
        )
        assert len(under_test._trusted) == 2

    def test_ignore_falsy_issuers_in_url_construction(self):
        under_test = OidcMultiIssuerValidator.from_auth_server_urls(
            audience=TEST_AUDIENCE,
            trusted_auth_server_urls=[TEST_PRIMARY_ISSUER, "", None],
        )
        assert len(under_test._trusted) == 1

        under_test = OidcMultiIssuerValidator.from_auth_server_urls(
            audience=TEST_AUDIENCE,
            trusted_auth_server_urls=[TEST_PRIMARY_ISSUER, TEST_SECONDARY_ISSUER, "", None],
        )
        assert len(under_test._trusted) == 2

    def test_ignore_falsy_issuers_in_configdict_construction(self):
        under_test = OidcMultiIssuerValidator.from_auth_server_configs(
            trusted_auth_server_configs=["", primary_issuer_config_dict, None],
        )
        assert len(under_test._trusted) == 1

        under_test = OidcMultiIssuerValidator.from_auth_server_configs(
            trusted_auth_server_configs=[
                "",
                primary_issuer_config_dict,
                secondary_issuer_config_dict,
                None,
            ],
        )
        assert len(under_test._trusted) == 2

    def test_reject_unknown_issuer(self):
        # QE TC15
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        access_token = untrusted_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__primary_and_secondary_validator()
        with pytest.raises(AuthException):
            local_validation, remote_validation = under_test.validate_access_token(
                access_token.access_token(), do_remote_revocation_check=False
            )

    def test_fail_bad_config_1(self):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = self.username(test_case_name)

        primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        with pytest.raises(AuthException):
            self.under_test__bad_client_1()

    def test_config_dict_construction_happy(self):
        test_conf_dict_1 = {
            "auth_server": TEST_PRIMARY_ISSUER,
            "scopes": ["test_scp0", "test_scp1"],
            "audiences": [TEST_AUDIENCE],
        }
        test_conf_dict_2 = {
            "auth_server": TEST_SECONDARY_ISSUER,
            "scopes": ["test_scp0", "test_scp1"],
            "audiences": [TEST_AUDIENCE],
        }
        under_test = OidcMultiIssuerValidator.from_auth_server_configs(
            trusted_auth_server_configs=[test_conf_dict_1, test_conf_dict_2]
        )
        assert len(under_test._trusted) == 2

    def test_config_dict_construction_sets_client_type(self):
        test_conf_dict = {
            "auth_server": TEST_PRIMARY_ISSUER,
            "scopes": ["test_scp0", "test_scp1"],
            "audiences": [TEST_AUDIENCE],
        }
        under_test = OidcMultiIssuerValidator.from_auth_server_configs(trusted_auth_server_configs=[test_conf_dict])
        assert len(under_test._trusted) == 1
        assert isinstance(under_test._trusted[TEST_PRIMARY_ISSUER].auth_client(), OidcClientValidatorAuthClient)
        assert (
            under_test._trusted[TEST_PRIMARY_ISSUER].auth_client()._oidc_client_config.issuer() == TEST_PRIMARY_ISSUER
        )

        test_conf_dict["client_type"] = "oidc_client_credentials_secret"
        test_conf_dict["client_id"] = "__test_dummy__"
        test_conf_dict["client_secret"] = "__test_dummy__"
        test_conf_dict["issuer"] = "__test_dummy__"
        under_test = OidcMultiIssuerValidator.from_auth_server_configs(trusted_auth_server_configs=[test_conf_dict])
        assert len(under_test._trusted) == 1
        assert isinstance(under_test._trusted["__test_dummy__"].auth_client(), ClientCredentialsClientSecretAuthClient)
        assert under_test._trusted["__test_dummy__"].auth_client()._oidc_client_config.issuer() == "__test_dummy__"

    @mock.patch("planet_auth.logging.auth_logger.AuthLogger.warning")
    @mock.patch("planet_auth.logging.auth_logger.AuthLogger.info")
    def test_primary_logs_as_info(self, mock_auth_logger_info, mock_auth_logger_warning):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = test_case_name + "_username"
        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__primary_and_secondary_validator(log_primary=True)
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False
        )

        assert mock_auth_logger_info.call_count == 1
        assert mock_auth_logger_warning.call_count == 0

    @mock.patch("planet_auth.logging.auth_logger.AuthLogger.warning")
    @mock.patch("planet_auth.logging.auth_logger.AuthLogger.info")
    def test_primary_logs_disabled(self, mock_auth_logger_info, mock_auth_logger_warning):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = test_case_name + "_username"
        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__primary_and_secondary_validator(log_primary=False)
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False
        )

        assert mock_auth_logger_info.call_count == 0
        assert mock_auth_logger_warning.call_count == 0

    @mock.patch("planet_auth.logging.auth_logger.AuthLogger.warning")
    @mock.patch("planet_auth.logging.auth_logger.AuthLogger.info")
    def test_primary_logs_by_default(self, mock_auth_logger_info, mock_auth_logger_warning):
        test_case_name = inspect.currentframe().f_code.co_name
        test_username = test_case_name + "_username"
        access_token = primary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        # access_token = secondary_issuer.login(username=test_username, extra_claims={"test_case": test_case_name})
        under_test = self.under_test__primary_and_secondary_validator()
        local_validation, remote_validation = under_test.validate_access_token(
            access_token.access_token(), do_remote_revocation_check=False
        )

        assert mock_auth_logger_info.call_count == 1
        assert mock_auth_logger_warning.call_count == 0



from planet_auth import Auth


def main():
    # The required files will vary by client type
    auth_client_config = {
        "client_type": "oidc_auth_code",
        "auth_server": "https://account-next.planet.com/oauth2/planet",
        "client_id": "your_client_id",
        "redirect_uri": "https://static.prod.planet-labs.com/authentication-handler/live/main/",
        "local_redirect_uri": "http://localhost:8080",
        "audiences": ["https://api.staging.planet-labs.com/"],
        "scopes": ["offline_access", "openid", "profile", "planet"],
    }
    auth_ctx = Auth.initialize_from_config_dict(client_config=auth_client_config)
    print(
        "Auth context initialized from in memory profile. Auth client class is {}.".format(
            auth_ctx.auth_client().__class__.__name__
        )
    )


if __name__ == "__main__":
    main()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import auto
from strenum import StrEnum


class AuthEvent(StrEnum):
    """
    Auth Events.

    Events exist to provide structure to log messages, with the
    initial use case in mind being building dashboards from log messages.
    Many events are error conditions, and these will overlap with
    exceptions, but not events indicate errors.
    """

    TRACE = auto()
    TOKEN_VALID = auto()
    TOKEN_INVALID = auto()
    TOKEN_INVALID_BAD_ALGORITHM = auto()
    TOKEN_INVALID_BAD_ISSUER = auto()
    TOKEN_INVALID_BAD_NONCE = auto()
    TOKEN_INVALID_BAD_SCOPE = auto()
    TOKEN_INVALID_BAD_SIGNING_KEY = auto()
    TOKEN_INVALID_EXPIRED = auto()
    TOKEN_INVALID_INTROSPECTION_REJECTION = auto()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import http.server

import importlib.resources  # nosemgrep

from http import HTTPStatus
from urllib.parse import urlparse, parse_qs, urlencode
from typing import Dict, List, Optional
from webbrowser import open_new

import planet_auth.logging.auth_logger
import planet_auth.oidc.util as oidc_util
from planet_auth.oidc import resources
from planet_auth.oidc.api_clients.api_client import OidcApiClientException

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()
DEFAULT_REDIRECT_LISTEN_PORT = 80
AUTH_TIMEOUT = 60


class AuthorizationApiException(OidcApiClientException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class _OidcSigninCallbackHandler(http.server.BaseHTTPRequestHandler):
    """
    HTTP Server callbacks to handle OAuth redirects.
    This handler expects to be invoked as a callback after user
    authentication. When called after successful login, the request will
    contain an auth code that can be used to obtain tokens. Parsing of the
    results is not handled here. It is the caller's responsibility.  All the
    callback does is pass the request along.
    """

    def __init__(self, request, address, server, do_logging: bool = False, response_body: str = None):
        self._do_logging = do_logging
        self._response_body_str = response_body
        super().__init__(request, address, server)

    def do_GET(self):
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.server.callback_raw_request_path = self.path
        self.wfile.write(bytes(self._response_body_str, "utf-8"))

    def log_message(self, *args, **kwargs) -> None:
        # do nothing to disable logging.  We let the caller control this so
        # they can hush the HTTP server logging regardless of the global log
        # level.
        if self._do_logging:
            return super().log_message(*args, **kwargs)


def _parse_authcode_from_callback(raw_request_path, expected_state):
    if not raw_request_path:
        raise AuthorizationApiException(message="Authorization callback was empty")

    auth_logger.debug(msg="Parsing callback request from authorization server" + raw_request_path)

    parsed_query_string = parse_qs(urlparse(raw_request_path).query)

    error_code = parsed_query_string.get("error")
    if error_code:
        # TODO: Can we unify this error parsing with that in the
        #       oidc api_client baseclass?
        error_description = parsed_query_string.get("error_description") or ["no error description"]
        raise AuthorizationApiException(
            message="Authorization error: {}: {}".format(error_code[0], error_description[0]),
            raw_response=raw_request_path,
        )

    state_array = parsed_query_string.get("state")
    state = None

    if state_array:
        state = state_array[0]
    if state != expected_state:
        raise AuthorizationApiException(
            message="Callback state did not match expected value. Expected: {}, Received: {}".format(
                expected_state, state
            ),
            raw_response=raw_request_path,
        )

    auth_code_array = parsed_query_string.get("code")
    auth_code = None
    if auth_code_array:
        auth_code = auth_code_array[0]
    if not auth_code:
        raise AuthorizationApiException(
            message="Failed to understand authorization callback. Callback request"
            " did not include an authorization code or a recognized error."
            " Raw callback request: " + raw_request_path,
            raw_response=raw_request_path,
        )

    return auth_code


class AuthorizationApiClient:
    """
    Low level API client for the OAuth Authorization endpoint. See
    RFC 6749 - The OAuth 2.0 Authorization Framework for protocol details.

    This is not a child of the OidcApiClient base class since, unlike
    most low level API clients, this class does not directly call the API.
    Rather, a web browser or other methods are used to interact with
    the API in most cases so that IDP redirects, MFA, and user
    interactive authentication can be performed.
    """

    def __init__(
        self, authorization_uri: str, authorization_callback_acknowledgement_response_body: Optional[str] = None
    ):
        """
        Create a new Authorization API Client.
        """
        self._authorization_uri = authorization_uri
        if authorization_callback_acknowledgement_response_body:
            self._authorization_callback_acknowledgement_response_body = (
                authorization_callback_acknowledgement_response_body
            )
        else:
            self._authorization_callback_acknowledgement_response_body = (
                importlib.resources.files(resources).joinpath("callback_acknowledgement.html").read_text("utf-8")
            )

    @staticmethod
    def prep_pkce_auth_payload(
        client_id: str,
        redirect_uri: str,
        requested_scopes: List[str],
        requested_audiences: List[str],
        pkce_code_challenge: str,
        extra: Dict,
    ) -> Dict:
        """
        Prepare the payload needed to make an authorization request to an
        OAuth authorization endpoint. This will usually be used to construct
        query parameters that are appended to the URL of the authorization
        endpoint, but some implementations are known to accept a POST rather
        than a GET with URL parameters.

        It should also be noted that this method does not prepare any sort
        of client authentication.  Depending on whether this is used in
        support of a confidential or non-confidential OAuth client, additional
        work may be required. The details of what may be required may
        vary depending on the auth server's demands of a confidential
        client; The auth server may require HTTP auth headers, or that
        additional assertions are added to the payload, or may require
        nothing at all.  This all depends on the auth server's configuration
        for the particular client.

        Parameters:
            client_id: The ID of the OAuth client requesting authorization.
            redirect_uri: The callback URI that should be used to return control
                to the client program after user authentication and authorization
                has been completed.
            requested_scopes: A list of strings specifying the scopes to
                request.
            requested_audiences: A list of strings specifying the audiences
                to request.
            pkce_code_challenge: PKCE challenge code.  The caller should generate
                this code along with a validation code that can be used to verify
                responses.
            extra: Dict of extra parameters to pass to the authorization endpoint.

        Returns:
            Returns a dict that is suitable for use in making an authorization request.

        Example:
            ```
            auth_url = AUTHORIZATION_ENDPOINT_URL + "?" + urlencode(prep_pkce_auth_payload(...))
            ```
        """
        if extra is None:
            extra = {}
        # "None" is pythonic, and does not mean anything to OAuth APIs.
        _extra = {k: v for k, v in extra.items() if v is not None}

        data = {
            **_extra,  # we do this first, because we want it to lose the race with explicit parameters. It is not intended for clobbering.
            "client_id": client_id,
            "response_type": "code",
            "redirect_uri": redirect_uri,
            "state": oidc_util.generate_nonce(8),
            "nonce": oidc_util.generate_nonce(32),
            "code_challenge": pkce_code_challenge,
            "code_challenge_method": "S256",
        }
        if requested_scopes:
            # Standard
            data["scope"] = " ".join(requested_scopes)
        if requested_audiences:
            # Used by Auth0. Others?
            data["audience"] = " ".join(requested_audiences)

        return data

    def authcode_from_pkce_auth_request_with_browser_and_callback_listener(
        self,
        client_id: str,
        redirect_uri: str,
        requested_scopes: List[str],
        requested_audiences: List[str],
        pkce_code_challenge: str,
        extra: Dict,
    ) -> str:
        """
        Request an authorization code by launching a web browser directed to the
        OAuth Authorization endpoint.  An HTTP listener will be set up to capture
        the browser redirect after interaction with the auth server and IDP is complete.

        Parameters:
            client_id: The ID of the OAuth client requesting authorization.
            redirect_uri: The callback URI that should be used to return control
                to the client program after user authentication and authorization
                has been completed.
            requested_scopes: A list of strings specifying the scopes to
                request.
            requested_audiences: A list of strings specifying the audiences
                to request.
            pkce_code_challenge: PKCE challenge code.  The caller should generate
                this code along with a validation code that can be used to verify
                responses.
            extra: Dict of extra parameters to pass to the authorization endpoint.

        Returns:
            Returns an auth code upon success. An exception should be raised upon error.
        """
        data = self.prep_pkce_auth_payload(
            client_id=client_id,
            redirect_uri=redirect_uri,
            requested_scopes=requested_scopes,
            requested_audiences=requested_audiences,
            pkce_code_challenge=pkce_code_challenge,
            extra=extra,
        )
        auth_request_uri = self._authorization_uri + "?" + urlencode(data)

        # HTTP server to catch the callback redirect from the browser
        # auth process
        auth_logger.debug(msg='Setting up listener for auth callback handler with URI "{}"'.format(redirect_uri))
        parsed_redirect_url = urlparse(redirect_uri)
        listen_port = parsed_redirect_url.port if parsed_redirect_url.port else DEFAULT_REDIRECT_LISTEN_PORT
        if (
            parsed_redirect_url.hostname
            and parsed_redirect_url.hostname.lower() != "localhost"
            and parsed_redirect_url.hostname != "127.0.0.1"
        ):
            raise AuthorizationApiException(
                message="Unexpected hostname in auth redirect URI. Expected"
                ' localhost URI, but received "{}"'.format(redirect_uri)
            )

        # Only bind to loopback! See
        # https://datatracker.ietf.org/doc/html/rfc8252#section-8.3
        http_server = http.server.HTTPServer(
            ("localhost", listen_port),
            lambda request, address, server: _OidcSigninCallbackHandler(
                request,
                address,
                server,
                do_logging=False,  # (logger.root.level <= logging.DEBUG),
                response_body=self._authorization_callback_acknowledgement_response_body,
            ),
        )
        http_server.timeout = AUTH_TIMEOUT

        # Don't kick off the browser until we are satisfied that the callback
        # handler is up and listening. UX team wanted this on the console
        print(
            "Opening browser for authorization and listening locally for"
            ' callback.\nIf this fails, retry with "no browser"'
            " option enabled.\n"
        )
        auth_logger.debug(msg='Opening browser with authorization URL : "{}"\n'.format(auth_request_uri))
        open_new(auth_request_uri)

        # Do we ever need to loop for multiple callbacks?
        # (No, this should never be needed.)
        http_server.handle_request()

        if hasattr(http_server, "callback_raw_request_path"):
            return _parse_authcode_from_callback(http_server.callback_raw_request_path, data["state"])
        else:
            raise AuthorizationApiException(
                message="Unknown error obtaining login tokens. No callback data was received."
            )

    def authcode_from_pkce_auth_request_with_tty_input(
        self,
        client_id: str,
        redirect_uri: str,
        requested_scopes: List[str],
        requested_audiences: List[str],
        pkce_code_challenge: str,
        extra: Dict,
    ) -> str:
        """
        Request an authorization code by prompting the user to visit a
        specific authorization URI, and then prompt the user for the
        resulting authorization code via a TTY prompt.  The redirect URI
        should point to an application that is capable of accepting the
        final callback from the authorization URI, and is ready to parse
        the response and provide the client with the authorization code
        so that it may be entered at the prompt.

        Parameters:
            client_id: The ID of the OAuth client requesting authorization.
            redirect_uri: The callback URI that should be used to return control
                to the client program after user authentication and authorization
                has been completed.
            requested_scopes: A list of strings specifying the scopes to
                request.
            requested_audiences: A list of strings specifying the audiences
                to request.
            pkce_code_challenge: PKCE challenge code.  The caller should generate
                this code along with a validation code that can be used to verify
                responses.
            extra: Dict of extra parameters to pass to the authorization endpoint.

        Returns:
            Returns an auth code upon success. An exception should be raised upon error.
        """
        # 1) Display URL for user to paste into browser.
        # 2) Wait for them to copy-paste the auth code URL.
        # X) The process of catching the redirect from the auth and parsing
        #    out the auth code is out of band of this code.
        data = self.prep_pkce_auth_payload(
            client_id=client_id,
            redirect_uri=redirect_uri,
            requested_scopes=requested_scopes,
            requested_audiences=requested_audiences,
            pkce_code_challenge=pkce_code_challenge,
            extra=extra,
        )
        auth_request_uri = self._authorization_uri + "?" + urlencode(data)
        print(
            "Please go to the following URL to proceed with login.\n"
            "After successful login, please provide the resulting"
            " authentication code.\n"
            "\n\t{}\n\n".format(auth_request_uri)
        )
        return getpass.getpass(prompt="Authentication code: ")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import unittest

from planet_auth_utils.builtins import Builtins, BuiltinsException
from planet_auth_utils.constants import EnvironmentVariables

from tests.test_planet_auth_utils.util import TestWithHomeDirProfiles
from tests.test_planet_auth_utils.unit.auth_utils.builtins_test_impl import BuiltinConfigurationProviderMockTestImpl


class TestBuiltInProfiles:
    def test_load_auth_client_config_blank(self):
        with pytest.raises(BuiltinsException):  # as be:
            Builtins.load_builtin_auth_client_config(None)
        with pytest.raises(BuiltinsException):  # as be:
            Builtins.load_builtin_auth_client_config("")

    def test_load_auth_client_config_custom_profile(self):
        with pytest.raises(BuiltinsException):  # as be:
            Builtins.load_builtin_auth_client_config("custom_non_builtin_profile")

    def test_builtin_profile_auth_client_config_dict_blank(self):
        with pytest.raises(BuiltinsException):  # as be:
            Builtins.builtin_profile_auth_client_config_dict(None)
        with pytest.raises(BuiltinsException):  # as be:
            Builtins.builtin_profile_auth_client_config_dict("")

    def test_builtin_all_profile_dicts_are_valid(self):
        for builtin_name in Builtins.builtin_profile_names():
            builtin_dict = Builtins.builtin_profile_auth_client_config_dict(builtin_name)
            assert builtin_dict is not None

    # TODO
    # def test_load_empty_builtins(self):
    #     assert isinstance(Builtins._builtin, EmptyBuiltinProfileConstants)

    # TODO
    # def test_load_custom_builtins(self):
    #     assert isinstance(Builtins._builtin, SomeCustomProviderClass)


class TestAuthClientContextInitHelpers(TestWithHomeDirProfiles, unittest.TestCase):
    def setUp(self):
        os.environ[EnvironmentVariables.AUTH_BUILTIN_PROVIDER] = (
            "tests.test_planet_auth_utils.unit.auth_utils.builtins_test_impl.BuiltinConfigurationProviderMockTestImpl"
        )
        Builtins._builtin = None  # Reset built-in state.
        self.setUp_testHomeDir()

    def test_deailas_profile(self):
        under_test_resolved = Builtins.dealias_builtin_profile(
            BuiltinConfigurationProviderMockTestImpl.BUILTIN_PROFILE_ALIAS_UTEST_ALIAS_1
        )
        self.assertEqual(under_test_resolved, BuiltinConfigurationProviderMockTestImpl.BUILTIN_PROFILE_NAME_UTEST_USER)

    def test_nested_deailas_profile(self):
        under_test_resolved = Builtins.dealias_builtin_profile(
            BuiltinConfigurationProviderMockTestImpl.BUILTIN_PROFILE_ALIAS_UTEST_ALIAS_2
        )
        self.assertEqual(under_test_resolved, BuiltinConfigurationProviderMockTestImpl.BUILTIN_PROFILE_NAME_UTEST_USER)

    def test_deailas_non_builtin(self):
        with pytest.raises(BuiltinsException):  # as be:
            Builtins.dealias_builtin_profile("some_user_defined_non_builtin_profile")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class EnvironmentVariables:
    """
    Environment Variables used in the planet_auth_utils packages
    """

    AUTH_API_KEY = "PL_API_KEY"
    """
    A literal Planet API key.
    """

    AUTH_CLIENT_ID = "PL_AUTH_CLIENT_ID"
    """
    Client ID for an OAuth service account
    """

    AUTH_CLIENT_SECRET = "PL_AUTH_CLIENT_SECRET"
    """
    Client Secret for an OAuth service account
    """

    AUTH_EXTRA = "PL_AUTH_EXTRA"
    """
    List of extra options.  Values should be formatted as <key>=<value>.
    Multiple options should be whitespace delimited.
    """

    AUTH_PROFILE = "PL_AUTH_PROFILE"
    """
    Name of a profile to use for auth client configuration.
    """

    AUTH_TOKEN = "PL_AUTH_TOKEN"
    """
    Literal token string.
    """

    AUTH_TOKEN_FILE = "PL_AUTH_TOKEN_FILE"
    """
    File path to use for storing tokens.
    """

    AUTH_ISSUER = "PL_AUTH_ISSUER"
    """
    Issuer to use when requesting or validating OAuth tokens.
    """

    AUTH_AUDIENCE = "PL_AUTH_AUDIENCE"
    """
    Audience to use when requesting or validating OAuth tokens.
    """

    AUTH_ORGANIZATION = "PL_AUTH_ORGANIZATION"
    """
    Organization to use when performing client authentication.
    Only used for some authentication mechanisms.
    """

    AUTH_PASSWORD = "PL_AUTH_PASSWORD"
    """
    Password to use when performing client authentication.
    Only used for some authentication mechanisms.
    """

    AUTH_SCOPE = "PL_AUTH_SCOPE"
    """
    List of scopes to request when requesting OAuth tokens.
    Multiple scopes should be whitespace delimited.
    """

    AUTH_USERNAME = "PL_AUTH_USERNAME"
    """
    Username to use when performing client authentication.
    Only used for some authentication mechanisms.
    """

    AUTH_LOGLEVEL = "PL_LOGLEVEL"
    """
    Specify the log level.
    """

    AUTH_BUILTIN_PROVIDER = "PL_AUTH_BUILTIN_CONFIG_PROVIDER"
    """
    Specify a python module and class that implement the BuiltinConfigurationProviderInterface abstract
    interface to provide the library and utility commands with some built-in configurations.
    """



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pathlib
import unittest

from requests.models import Response
from unittest import mock

from planet_auth.auth_client import AuthClientConfigException
from planet_auth.planet_legacy.auth_client import (
    PlanetLegacyAuthClient,
    PlanetLegacyAuthClientConfig,
    PlanetLegacyAuthClientException,
)
from planet_auth.planet_legacy.legacy_api_key import FileBackedPlanetLegacyApiKey
from planet_auth.planet_legacy.request_authenticator import PlanetLegacyRequestAuthenticator
from tests.test_planet_auth.unit.auth.util import TestTokenBuilder
from tests.test_planet_auth.util import load_rsa_private_key, tdata_resource_file_path

TEST_MOCK_API_KEY = "PLAK_MockApiKey"
TEST_MOCK_API_KEY2 = "PLAK_MockApiKey2"
TEST_AUTH_ENDPOINT = "https://blackhole.unittest.planet.com/legacy_auth"
TEST_BAD_RESPONSE = {"some_key": "some bogus response payload"}
TEST_TOKEN_TTL = 60

TEST_SIGNING_KEY = load_rsa_private_key(tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem"))

TOKEN_BUILDER = TestTokenBuilder(
    issuer="__test_token_issuer__",
    audience="__test_token_audience__",
    signing_key_file=tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem"),
)


def generate_response_struct(token_str):
    response_struct = {"token": token_str}
    return response_struct


def mock_response_valid(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(
        json.dumps(
            generate_response_struct(
                TOKEN_BUILDER.generate_legacy_token(api_key=TEST_MOCK_API_KEY, ttl=TEST_TOKEN_TTL)
            )
        )
    )
    return response


def mock_response_http_error(request_url, **kwargs):
    response = Response()
    response.status_code = 401
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps({"error": "access denied"}))
    return response


def mock_response_empty_payload(request_url, **kwargs):
    response = Response()
    response.status_code = 202
    return response


def mock_response_bad_content_type(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "text/plain"
    response._content = str.encode("Plain sting response")
    return response


def mock_response_no_token_in_payload(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps({"foo": "this test payload lacks a token"}))
    return response


def mock_response_bad_token(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(
        json.dumps(generate_response_struct(TOKEN_BUILDER.generate_legacy_token(api_key=None, ttl=TEST_TOKEN_TTL)))
    )
    return response


def mocked_getpass_password(**kwargs):
    return "mock_getpass_password"


def mocked_input_username(prompt, **kwargs):
    return "mock_input_username"


class TestLecaycApiAuthClientConfig(unittest.TestCase):
    def test_check(self):
        under_test = PlanetLegacyAuthClientConfig(api_key=TEST_MOCK_API_KEY, legacy_auth_endpoint=TEST_AUTH_ENDPOINT)
        under_test.check()  # should not throw

        under_test = PlanetLegacyAuthClientConfig(api_key=TEST_MOCK_API_KEY)
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


class TestLegacyApiAuthClient(unittest.TestCase):
    def setUp(self):
        self.under_test_no_apikey_in_conf_in_memory = PlanetLegacyAuthClient(
            PlanetLegacyAuthClientConfig(legacy_auth_endpoint=TEST_AUTH_ENDPOINT)
        )
        self.under_test_with_apikey_in_memory = PlanetLegacyAuthClient(
            PlanetLegacyAuthClientConfig(legacy_auth_endpoint=TEST_AUTH_ENDPOINT, api_key=TEST_MOCK_API_KEY2)
        )
        self.under_test_with_apikey_from_file_conf_path = tdata_resource_file_path(
            "auth_client_configs/utest/planet_legacy_with_apikey.json"
        )
        self.under_test_with_apikey_from_file = PlanetLegacyAuthClient(
            PlanetLegacyAuthClientConfig(file_path=self.under_test_with_apikey_from_file_conf_path)
        )

    @mock.patch("requests.sessions.Session.post", side_effect=mock_response_valid)
    def test_login_success_direct_input(self, mock1):
        test_result = self.under_test_no_apikey_in_conf_in_memory.login(username="test_user", password="test_password")
        self.assertIsInstance(test_result, FileBackedPlanetLegacyApiKey)
        self.assertEqual(TEST_MOCK_API_KEY, test_result.legacy_api_key())

    @mock.patch("requests.sessions.Session.post", side_effect=mock_response_valid)
    @mock.patch("getpass.getpass", mocked_getpass_password)
    @mock.patch("builtins.input", mocked_input_username)
    def test_login_success_user_prompt(self, mock1):
        test_result = self.under_test_no_apikey_in_conf_in_memory.login(allow_tty_prompt=True)
        self.assertIsInstance(test_result, FileBackedPlanetLegacyApiKey)
        self.assertEqual(TEST_MOCK_API_KEY, test_result.legacy_api_key())

    def test_login_failure_username_input_needed(self):
        with self.assertRaises(PlanetLegacyAuthClientException):
            self.under_test_no_apikey_in_conf_in_memory.login(
                username=None, password="test_password", allow_tty_prompt=False
            )

    def test_login_failure_password_input_needed(self):
        with self.assertRaises(PlanetLegacyAuthClientException):
            self.under_test_no_apikey_in_conf_in_memory.login(
                username="test_user", password=None, allow_tty_prompt=False
            )

    @mock.patch("requests.sessions.Session.post", side_effect=mock_response_http_error)
    def test_login_http_error(self, mock1):
        with self.assertRaises(PlanetLegacyAuthClientException):
            self.under_test_no_apikey_in_conf_in_memory.login(username="test_user", password="test_password")

    @mock.patch("requests.sessions.Session.post", side_effect=mock_response_empty_payload)
    def test_login_bad_response_no_payload(self, mock1):
        with self.assertRaises(PlanetLegacyAuthClientException):
            self.under_test_no_apikey_in_conf_in_memory.login(username="test_user", password="test_password")

    @mock.patch("requests.sessions.Session.post", side_effect=mock_response_bad_content_type)
    def test_login_bad_response_not_json(self, mock1):
        with self.assertRaises(PlanetLegacyAuthClientException):
            self.under_test_no_apikey_in_conf_in_memory.login(username="test_user", password="test_password")

    @mock.patch("requests.sessions.Session.post", side_effect=mock_response_no_token_in_payload)
    def test_login_bad_response_no_token(self, mock1):
        with self.assertRaises(PlanetLegacyAuthClientException):
            self.under_test_no_apikey_in_conf_in_memory.login(username="test_user", password="test_password")

    @mock.patch("requests.sessions.Session.post", side_effect=mock_response_bad_token)
    def test_login_bad_response_token_lacks_api_key(self, mock1):
        with self.assertRaises(PlanetLegacyAuthClientException):
            self.under_test_no_apikey_in_conf_in_memory.login(username="test_user", password="test_password")

    def test_default_request_authenticator_when_login(self):
        test_result = self.under_test_no_apikey_in_conf_in_memory.default_request_authenticator(
            credential=pathlib.Path("/test/token.json")
        )
        self.assertIsInstance(test_result, PlanetLegacyRequestAuthenticator)
        self.assertEqual(pathlib.Path("/test/token.json"), test_result._credential.path())

    def test_default_request_authenticator_when_apikey_in_client_config_from_file(self):
        test_result = self.under_test_with_apikey_from_file.default_request_authenticator(
            credential=pathlib.Path("/test/token.json")
        )
        self.assertIsInstance(test_result, PlanetLegacyRequestAuthenticator)
        self.assertEqual(PlanetLegacyRequestAuthenticator.TOKEN_PREFIX, test_result._token_prefix)
        self.assertEqual("__utest_legacy_api_key_in_client_config_file__", test_result._credential.legacy_api_key())
        self.assertIsNone(test_result._token_body)  # Loaded JIT
        test_result.pre_request_hook()
        self.assertEqual("__utest_legacy_api_key_in_client_config_file__", test_result._token_body)
        self.assertEqual(self.under_test_with_apikey_from_file_conf_path, test_result._credential.path())

    def test_default_request_authenticator_when_apikey_in_client_config_in_memory(self):
        test_result = self.under_test_with_apikey_in_memory.default_request_authenticator(
            credential=pathlib.Path("/test/token.json")
        )
        self.assertIsInstance(test_result, PlanetLegacyRequestAuthenticator)
        self.assertEqual(PlanetLegacyRequestAuthenticator.TOKEN_PREFIX, test_result._token_prefix)
        self.assertEqual(TEST_MOCK_API_KEY2, test_result._credential.legacy_api_key())
        self.assertIsNone(test_result._token_body)  # Loaded JIT
        test_result.pre_request_hook()
        self.assertEqual(TEST_MOCK_API_KEY2, test_result._token_body)
        self.assertEqual(None, test_result._credential.path())

    def test_default_request_authenticator_given_literal_credential_object_instead_of_path(self):
        literal_credential = FileBackedPlanetLegacyApiKey(api_key="literal_utest_api_credential")
        test_result = self.under_test_no_apikey_in_conf_in_memory.default_request_authenticator(
            credential=literal_credential
        )
        test_result.pre_request_hook()
        self.assertEqual("literal_utest_api_credential", test_result._token_body)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import logging
import importlib.metadata
import sys
import time

from planet_auth import Auth, AuthException, setStructuredLogging, ObjectStorageProvider, ObjectStorageProvider_KeyType
from planet_auth.constants import USER_CONFIG_FILE

from planet_auth_utils.plauth_factory import PlanetAuthFactory
from planet_auth_utils.profile import Profile

from .options import (
    opt_organization,
    opt_project,
    opt_profile,
    opt_client_id,
    opt_client_secret,
    opt_api_key,
    opt_username,
    opt_password,
    opt_loglevel,
    opt_open_browser,
    opt_show_qr_code,
    opt_sops,
    opt_audience,
    opt_scope,
    opt_yes_no,
)
from .oauth_cmd import cmd_oauth
from .planet_legacy_auth_cmd import cmd_pllegacy
from .profile_cmd import cmd_profile
from .jwt_cmd import cmd_jwt
from .util import recast_exceptions_to_click, post_login_cmd_helper


@click.group("plauth", invoke_without_command=True, help="Planet authentication utility")
@opt_loglevel
@opt_profile
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_plauth(ctx, loglevel, auth_profile):
    """
    Planet Auth Utility commands
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        sys.exit(0)

    # cli_logger = logging.getLogger("plauth-logger")
    # setPyLoggerForAuthLogger(cli_logger)
    setStructuredLogging(nested_key=None)
    logging.basicConfig(level=loglevel)

    ctx.ensure_object(dict)

    ctx.obj["AUTH"] = PlanetAuthFactory.initialize_auth_client_context(
        auth_profile_opt=auth_profile,
        # token_file_opt=token_file,
    )


@click.group("plauth", invoke_without_command=True, help="Embedded PLAuth advanced authentication utility")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_plauth_embedded(ctx):
    """
    Planet Auth Utility commands

    Embeddable version of the Planet Auth Client root command.
    The embedded command differs from the stand-alone command in that it
    expects the context to be instantiated and options to be handled by
    the parent command.  See [planet_auth_utils.PlanetAuthFactory.initialize_auth_client_context][]
    for user-friendly auth client context initialization.

    See [examples](/examples/#embedding-the-click-auth-command).
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        sys.exit(0)

    ctx.ensure_object(dict)

    if not isinstance(ctx.obj.get("AUTH"), Auth):
        raise click.ClickException(
            "INTERNAL ERROR:"
            "  The Auth context is expected to be created by the caller when using the embedded plauth command."
            "  This is a programming error, and must be fixed by the developer."
            "  See developer documentation."
        )


@cmd_plauth.command("version")
def cmd_plauth_version():
    """
    Show the version of planet auth components.
    """

    def _pkg_display_version(pkg_name):
        try:
            return importlib.metadata.version(pkg_name)
        except importlib.metadata.PackageNotFoundError:
            return "N/A"

    # Well known packages with built-in profile configs we commonly use.
    print(f"planet-auth : {_pkg_display_version('planet-auth')}")
    print(f"planet-auth-config : {_pkg_display_version('planet-auth-config')}")
    print(f"planet : {_pkg_display_version('planet')}")


@cmd_plauth.command("reset")
def cmd_plauth_reset():
    """
    Reset saved auth state.

    Old auth state is not deleted. It is moved aside and preserved.
    """
    save_tag = time.strftime("%Y-%m-%d-%H%M%S")

    # The CLI only supports the default storage provider right now.
    storage_provider = ObjectStorageProvider._default_storage_provider()

    user_conf_objkey = ObjectStorageProvider_KeyType(USER_CONFIG_FILE)
    if storage_provider.obj_exists(user_conf_objkey):
        user_conf_objkey_offname = ObjectStorageProvider_KeyType(USER_CONFIG_FILE + f"-{save_tag}")
        storage_provider.obj_rename(user_conf_objkey, user_conf_objkey_offname)

    profile_dir_objkey = ObjectStorageProvider_KeyType(Profile.profile_root())
    if storage_provider.obj_exists(profile_dir_objkey):
        profile_dir_objkey_offname = ObjectStorageProvider_KeyType(Profile.profile_root().name + f"-{save_tag}")
        storage_provider.obj_rename(profile_dir_objkey, profile_dir_objkey_offname)


@cmd_plauth.command("login")
@opt_open_browser
@opt_show_qr_code
@opt_scope
@opt_audience()
@opt_organization
@opt_project
@opt_profile
@opt_client_id
@opt_client_secret
@opt_api_key
@opt_username()
@opt_password()
@opt_sops
@opt_yes_no
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError, PermissionError)
def cmd_plauth_login(
    ctx,
    scope,
    audience,
    open_browser,
    show_qr_code,
    organization,
    project,
    auth_profile,
    auth_client_id,
    auth_client_secret,
    auth_api_key,
    username,
    password,
    sops,
    yes,
):
    """
    Perform an initial login, obtain user authorization, and save access
    tokens for the selected authentication profile.  The specific process
    used depends on the selected options and authentication profile.
    """
    extra = {}
    if project:
        # Planet Labs OAuth extension to request a token for a particular project
        extra["project_id"] = project
    if organization:
        # Used by Auth0's OAuth implementation to support their concept of selecting
        # a particular organization at login when the user belongs to more than one.
        extra["organization"] = organization

    # Arguments to login commands may imply an override to the default/root
    # command auth provider in a way that is different from what we expect
    # in most non-root commands.
    # root_cmd_auth_context = ctx.obj["AUTH"]
    override_auth_context = PlanetAuthFactory.initialize_auth_client_context(
        auth_profile_opt=auth_profile,
        auth_client_id_opt=auth_client_id,
        auth_client_secret_opt=auth_client_secret,
        auth_api_key_opt=auth_api_key,
        # auth_username_opt=auth_username,
        # auth_password_opt=auth_password,
    )

    print(f"Logging in with authentication profile {override_auth_context.profile_name()}...")
    _ = override_auth_context.login(
        requested_scopes=scope,
        requested_audiences=audience,
        allow_open_browser=open_browser,
        allow_tty_prompt=True,
        display_qr_code=show_qr_code,
        username=username,
        password=password,
        client_id=auth_client_id,
        client_secret=auth_client_secret,
        extra=extra,
    )
    print("Login succeeded.")  # Errors should throw.

    post_login_cmd_helper(override_auth_context=override_auth_context, use_sops=sops, prompt_pre_selection=yes)


cmd_plauth.add_command(cmd_oauth)
cmd_plauth.add_command(cmd_pllegacy)
cmd_plauth.add_command(cmd_profile)
cmd_plauth.add_command(cmd_jwt)

cmd_plauth_embedded.add_command(cmd_oauth)
cmd_plauth_embedded.add_command(cmd_pllegacy)
cmd_plauth_embedded.add_command(cmd_profile)
cmd_plauth_embedded.add_command(cmd_jwt)
cmd_plauth_embedded.add_command(cmd_plauth_login)
cmd_plauth_embedded.add_command(cmd_plauth_version)


if __name__ == "__main__":
    cmd_plauth()  # pylint: disable=E1120



# Copyright 2025 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
import pytest

from abc import ABC
from unittest import mock
from requests import Response

from planet_auth.constants import TOKEN_FILE_PLAIN, AUTH_CONFIG_FILE_PLAIN
from planet_auth_utils.plauth_factory import PlanetAuthFactory
from planet_auth_utils.profile import Profile

from tests.test_planet_auth.unit.auth.util import MockObjectStorageProvider, TestTokenBuilder


# This isn't the best unit test of Auth.  Rather, this tests how several
# things come together to make the request_authenticator_is_ready() work as
# expected in the Auth class.  We run this test for a number of common
# auth clients and authenticators.  This probably points to a code smell
# that this part of the library isn't that cleanly encapsulated.
# It's worth investigating if the Auth class should be some form of GSSAPI
# binding.

TEST_AUTH_SERVER = "https://login-utest.planet.com/"
TEST_AUDIENCE = "https://utest.planet.com/"


_pubkey, _mock_token_builder = TestTokenBuilder.test_token_builder_factory(keypair_name="keypair1")


def mocked_get_password(**kwargs):
    return "_mock_authcode_"


def _mock_oidc_credential_data() -> dict:
    # Data that is the same as what we expect in FileBackedOidcCredential.data
    return {
        "access_token": _mock_token_builder.construct_oidc_access_token_rfc8693(
            username="utest_user", requested_scopes=["planet", "offline_access"], ttl=100
        ),
        "refresh_token": "__FAKE_UTEST_REFRESH_TOKEN__",
        "expires_in": 100,
    }


def mock_oauth_http_svc(request_url, **kwargs):
    mock_http_data = {
        f"{TEST_AUTH_SERVER}.well-known/openid-configuration": {
            "issuer": f"{TEST_AUTH_SERVER}",
            "authorization_endpoint": f"{TEST_AUTH_SERVER}oauth/authorize",
            "token_endpoint": f"{TEST_AUTH_SERVER}oauth/token",
            "device_authorization_endpoint": f"{TEST_AUTH_SERVER}oauth/device/code",
        },
        f"{TEST_AUTH_SERVER}oauth/device/code": {
            "device_code": "mock_device_code",
            "user_code": "mock_user_code",
            "verification_uri": f"{TEST_AUTH_SERVER}verify",
            "expires_in": 100,
        },
        f"{TEST_AUTH_SERVER}oauth/authorize": {},
        f"{TEST_AUTH_SERVER}oauth/token": _mock_oidc_credential_data(),
    }
    response = Response()
    response.headers["content-type"] = "application/json"
    if request_url in mock_http_data:
        response.status_code = 200
        response._content = str.encode(json.dumps(mock_http_data[request_url]))
    else:
        response.status_code = 404

    return response


class IsInitializedTestBaseForOAuth2Client(ABC):
    def _profilePath(self, profile_name, file_name):
        return Profile.get_profile_file_path(profile=profile_name, filename=file_name)

    def _initMockStorage(self):
        self.test_client_data_in_storage = {
            "storage-utest-m2m-with-credential": {
                "client-config": {
                    "client_type": "oidc_client_credentials_secret",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_1__",
                    "client_secret": "__UTEST_CLIENT_SECRET__",
                },
                "saved-token": _mock_oidc_credential_data(),
            },
            "storage-utest-m2m-without-credential": {
                "client-config": {
                    "client_type": "oidc_client_credentials_secret",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_2__",
                    "client_secret": "__UTEST_CLIENT_SECRET__",
                },
            },
            "storage-utest-device-code-with-credential": {
                "client-config": {
                    "client_type": "oidc_device_code",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_3__",
                    "scopes": [
                        "planet",
                        "offline_access",
                    ],
                },
                "saved-token": _mock_oidc_credential_data(),
            },
            "storage-utest-device-code-without-credential": {
                "client-config": {
                    "client_type": "oidc_device_code",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_4__",
                    "scopes": [
                        "planet",
                        "offline_access",
                    ],
                },
            },
            "storage-utest-auth-code-with-credential": {
                "client-config": {
                    "client_type": "oidc_auth_code",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_5__",
                    "redirect_uri": "http://localhost:8080/",
                    "scopes": [
                        "planet",
                        "offline_access",
                    ],
                },
                "saved-token": _mock_oidc_credential_data(),
            },
            "storage-utest-auth-code-without-credential": {
                "client-config": {
                    "client_type": "oidc_auth_code",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_6__",
                    "redirect_uri": "http://localhost:8080/",
                    "scopes": [
                        "planet",
                        "offline_access",
                    ],
                },
            },
        }

        self.test_client_data_in_memory = {
            "mem-utest-m2m-without-credential": {
                "client-config": {
                    "client_type": "oidc_client_credentials_secret",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_102__",
                    "client_secret": "__UTEST_CLIENT_SECRET__",
                },
            },
            "mem-utest-device-code-without-credential": {
                "client-config": {
                    "client_type": "oidc_device_code",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_104__",
                    "scopes": [
                        "planet",
                        "offline_access",
                    ],
                },
            },
            "mem-utest-auth-code-without-credential": {
                "client-config": {
                    "client_type": "oidc_auth_code",
                    "auth_server": f"{TEST_AUTH_SERVER}",
                    "audiences": [f"{TEST_AUDIENCE}"],
                    "client_id": "__UTEST_CLIENT_ID_106__",
                    "redirect_uri": "http://localhost:8080/",
                    "scopes": [
                        "planet",
                        "offline_access",
                    ],
                },
            },
        }

        initial_mock_storage_initial_state = {}
        for profile_name in self.test_client_data_in_storage:
            client_data = self.test_client_data_in_storage[profile_name]
            client_conf_path = self._profilePath(profile_name=profile_name, file_name=AUTH_CONFIG_FILE_PLAIN)
            token_path = self._profilePath(profile_name=profile_name, file_name=TOKEN_FILE_PLAIN)
            if "client-config" in client_data:
                initial_mock_storage_initial_state[client_conf_path] = client_data["client-config"]
            if "saved-token" in client_data:
                initial_mock_storage_initial_state[token_path] = client_data["saved-token"]

        self.mock_storage_provider = MockObjectStorageProvider(initial_mock_storage=initial_mock_storage_initial_state)

    def setUp(self):
        self._initMockStorage()


class TestAuthIsInitialized_authCodeAuthClient(unittest.TestCase, IsInitializedTestBaseForOAuth2Client):
    def setUp(self):
        super()._initMockStorage()

    def test_session_is_initialized__in_memory__without_initial_credential_data(self):
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_memory["mem-utest-auth-code-without-credential"]["client-config"],
            storage_provider=self.mock_storage_provider,
            profile_name="some-random-utest-name",
            save_profile_config=False,
            save_token_file=False,
        )
        self.assertFalse(under_test.request_authenticator_is_ready())

    def test_session_is_initialized_by_existing_credential__in_memory__with_initial_credential_data(self):
        # Not expected to be a common case.  If you have a saved credential, you should be using storage.
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_memory["mem-utest-auth-code-without-credential"]["client-config"],
            storage_provider=self.mock_storage_provider,
            profile_name="some-random-utest-name",
            save_profile_config=False,
            save_token_file=False,
            initial_token_data=_mock_oidc_credential_data(),
        )
        self.assertTrue(under_test.request_authenticator_is_ready())

    def test_session_is_initialized_by_existing_credential__with_storage__with_stored_credential_data(self):
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_storage["storage-utest-auth-code-with-credential"]["client-config"],
            storage_provider=self.mock_storage_provider,
            profile_name="storage-utest-auth-code-with-credential",
            save_profile_config=True,
            save_token_file=True,
            initial_token_data=None,
        )
        self.assertTrue(under_test.request_authenticator_is_ready())

    @mock.patch("requests.sessions.Session.post", side_effect=mock_oauth_http_svc)
    @mock.patch("requests.sessions.Session.get", side_effect=mock_oauth_http_svc)
    @mock.patch("getpass.getpass", mocked_get_password)
    def test_session_is_initialized_by_login__in_memory__login_causes_session_to_be_initialized(
        self, mock_http_get, mock_http_post
    ):
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_memory["mem-utest-auth-code-without-credential"]["client-config"],
            storage_provider=self.mock_storage_provider,
            profile_name="some-random-utest-name",
            save_profile_config=False,
            save_token_file=False,
        )
        self.assertFalse(under_test.request_authenticator_is_ready())
        under_test.login(allow_tty_prompt=True, allow_open_browser=False)
        self.assertTrue(under_test.request_authenticator_is_ready())

    @pytest.mark.xfail(reason="TODO: Custom storage providers used by this test not yet supported for client profiles")
    @mock.patch("requests.sessions.Session.post", side_effect=mock_oauth_http_svc)
    @mock.patch("requests.sessions.Session.get", side_effect=mock_oauth_http_svc)
    def test_session_is_initialized_by_login__with_storage__profile_from_storage(self, mock_http_get, mock_http_post):
        under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_profile_opt="storage-utest-auth-code-without-credential",
            save_profile_config=True,
            save_token_file=True,
            # storage_provider=self.mock_storage_provider,
        )
        self.assertFalse(under_test.request_authenticator_is_ready())
        under_test.login(allow_tty_prompt=True, allow_open_browser=False)
        self.assertTrue(under_test.request_authenticator_is_ready())


class TestAuthIsInitialized_deviceCodeAuthClient(unittest.TestCase, IsInitializedTestBaseForOAuth2Client):
    def setUp(self):
        super()._initMockStorage()

    def test_session_is_initialized__in_memory__without_initial_credential_data(self):
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_memory["mem-utest-device-code-without-credential"]["client-config"],
            storage_provider=self.mock_storage_provider,
            profile_name="some-random-utest-name",
            save_profile_config=False,
            save_token_file=False,
        )
        self.assertFalse(under_test.request_authenticator_is_ready())

    def test_session_is_initialized_by_existing_credential__in_memory__with_initial_credential_data(self):
        # Not expected to be a common case.  If you have a saved credential, you should be using storage.
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_memory["mem-utest-device-code-without-credential"]["client-config"],
            storage_provider=self.mock_storage_provider,
            profile_name="some-random-utest-name",
            save_profile_config=False,
            save_token_file=False,
            initial_token_data=_mock_oidc_credential_data(),
        )
        self.assertTrue(under_test.request_authenticator_is_ready())

    def test_session_is_initialized_by_existing_credential__with_storage__with_stored_credential_data(self):
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_storage["storage-utest-device-code-with-credential"][
                "client-config"
            ],
            storage_provider=self.mock_storage_provider,
            profile_name="storage-utest-device-code-with-credential",
            save_profile_config=True,
            save_token_file=True,
            initial_token_data=None,
        )
        self.assertTrue(under_test.request_authenticator_is_ready())

    @mock.patch("requests.sessions.Session.post", side_effect=mock_oauth_http_svc)
    @mock.patch("requests.sessions.Session.get", side_effect=mock_oauth_http_svc)
    def test_session_is_initialized_by_login__in_memory__login_causes_session_to_be_initialized(
        self, mock_http_get, mock_http_post
    ):
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_memory["mem-utest-device-code-without-credential"]["client-config"],
            storage_provider=self.mock_storage_provider,
            profile_name="some-random-utest-name",
            save_profile_config=False,
            save_token_file=False,
        )
        self.assertFalse(under_test.request_authenticator_is_ready())
        device_login_initiate = under_test.device_login_initiate(allow_tty_prompt=True, allow_open_browser=False)
        under_test.device_login_complete(device_login_initiate)
        self.assertTrue(under_test.request_authenticator_is_ready())

    @pytest.mark.xfail(reason="TODO: Custom storage providers used by this test not yet supported for client profiles")
    @mock.patch("requests.sessions.Session.post", side_effect=mock_oauth_http_svc)
    @mock.patch("requests.sessions.Session.get", side_effect=mock_oauth_http_svc)
    def test_session_is_initialized_by_login__with_storage__profile_from_storage(self, mock_http_get, mock_http_post):
        under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_profile_opt="storage-utest-device-code-without-credential",
            save_profile_config=True,
            save_token_file=True,
            # storage_provider=self.mock_storage_provider,
        )
        self.assertFalse(under_test.request_authenticator_is_ready())
        under_test.login(allow_tty_prompt=True, allow_open_browser=False)
        self.assertTrue(under_test.request_authenticator_is_ready())


class TestAuthIsInitialized_m2mAuthClient(unittest.TestCase, IsInitializedTestBaseForOAuth2Client):
    def setUp(self):
        super()._initMockStorage()

    def test_session_is_initialized_by_client_config__in_memory(self):
        under_test = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config=self.test_client_data_in_memory["mem-utest-m2m-without-credential"]["client-config"],
            save_token_file=False,
            save_profile_config=False,
            profile_name="some-random-utest-name",
            # storage_provider=self.mock_storage_provider, # Should not be used or matter in this case.
        )
        self.assertTrue(under_test.request_authenticator_is_ready())

    @pytest.mark.xfail(reason="TODO: Custom storage providers used by this test not yet supported for client profiles")
    def test_session_is_initialized_by_client_config__from_storage_profile(self):
        under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_profile_opt="storage-utest-m2m-without-credential"
        )
        self.assertTrue(under_test.request_authenticator_is_ready())



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import pytest
import unittest
from unittest import mock

from planet_auth import RefreshOrReloginOidcTokenRequestAuthenticator, FileBackedOidcCredential, FileBackedApiKey
from planet_auth.auth_client import AuthClientConfigException
from planet_auth.oidc.auth_clients.resource_owner_flow import (
    ResourceOwnerClientConfig,
    ResourceOwnerAuthClient,
    ResourceOwnerWithClientSecretClientConfig,
    ResourceOwnerWithClientSecretAuthClient,
    ResourceOwnerWithPubKeyClientConfig,
    ResourceOwnerWithPubKeyAuthClient,
    ResourceOwnerAuthClientException,
)
from tests.test_planet_auth.unit.auth.util import StubOidcClientConfig, StubOidcAuthClient
from tests.test_planet_auth.util import tdata_resource_file_path

TEST_AUTH_SERVER = "https://blackhole.unittest.planet.com/fake_authserver"
TEST_CLIENT_ID = "fake_test_client_id"
TEST_CLIENT_SECRET = "fake_client_secret"
TEST_USER_NAME = "fake_username"
TEST_USER_PASSWORD = "fake_password"
TEST_TOKEN_SAVE_FILE_PATH = pathlib.Path("/test/token.json")
MOCK_TOKEN = {"token_type": "Bearer", "expires_in": 3600, "access_token": "__mock_access_token__", "scope": "planet"}

# Get some tokens that look more real than the mock token for some tests.
TEST_SIGNING_KEY_FILE = tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem")
TEST_SIGNING_PUBKEY_FILE = tdata_resource_file_path("keys/keypair1_pub_jwk.json")
STUB_AUTH_CLIENT_CONFIG = StubOidcClientConfig(
    auth_server=TEST_AUTH_SERVER,
    stub_authority_ttl=3600,
    stub_authority_access_token_audience="utest_audience",
    stub_authority_signing_key_file=TEST_SIGNING_KEY_FILE,
    stub_authority_pub_key_file=TEST_SIGNING_PUBKEY_FILE,
    scopes=["offline_access", "profile", "openid", "test_scope_1", "test_scope_2"],
)
STUB_AUTH_CLIENT = StubOidcAuthClient(STUB_AUTH_CLIENT_CONFIG)


@pytest.mark.skip("Test disabled (and probably should be deleted). No validation logic to test in the baseclass.")
class ResourceOwnerClientConfigTest(unittest.TestCase):
    def test_username_required(self):
        # No exception
        under_test = ResourceOwnerClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
        )
        under_test.check()

        under_test = ResourceOwnerClientConfig(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, username=None, password=TEST_USER_PASSWORD
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.check()

    def test_password_required(self):
        # No exception
        under_test = ResourceOwnerClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_secret=TEST_CLIENT_SECRET,
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
        )
        under_test.check()

        under_test = ResourceOwnerClientConfig(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, username=TEST_USER_NAME, password=None
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


def mocked_tokenapi_password(
    obj_self,
    client_id,
    username,
    password,
    requested_scopes,
    requested_audiences,
    auth_enricher,
    **kwargs,
):
    return MOCK_TOKEN


def mocked_getpass_password(**kwargs):
    return "mock_getpass_password"


def mocked_input_username(prompt, **kwargs):
    return "mock_input_username"


class ResourceOwnerNoClientAuthTest(unittest.TestCase):
    # Also doing base class tests here.
    def setUp(self):
        self.under_test_with_username_password = ResourceOwnerAuthClient(
            ResourceOwnerClientConfig(
                auth_server=TEST_AUTH_SERVER,
                client_id=TEST_CLIENT_ID,
                username=TEST_USER_NAME,
                password=TEST_USER_PASSWORD,
                token_endpoint=TEST_AUTH_SERVER + "/token",
            )
        )
        self.under_test_without_username_password = ResourceOwnerAuthClient(
            ResourceOwnerClientConfig(
                auth_server=TEST_AUTH_SERVER,
                client_id=TEST_CLIENT_ID,
                username=None,
                password=None,
                token_endpoint=TEST_AUTH_SERVER + "/token",
            )
        )

    @mock.patch(
        "planet_auth.oidc.api_clients.token_api_client.TokenApiClient.get_token_from_password",
        mocked_tokenapi_password,
    )
    def test_login(self):
        test_result = self.under_test_with_username_password.login()
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        self.assertEqual(MOCK_TOKEN, test_result.data())

        # again with override scopes, but since the response is mocked
        # there is nothing different to check in the result data.
        # Same goes for audiences.
        test_result = self.under_test_with_username_password.login(
            requested_scopes=["override1"], requested_audiences=["req_aud1"]
        )
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        # self.assertEqual(MOCK_TOKEN, test_result.data())

    @mock.patch(
        "planet_auth.oidc.api_clients.token_api_client.TokenApiClient.get_token_from_password",
        mocked_tokenapi_password,
    )
    @mock.patch("getpass.getpass", side_effect=mocked_getpass_password)
    @mock.patch("builtins.input", side_effect=mocked_input_username)
    def test_login_interactive(self, mock1, mock2):
        self.under_test_without_username_password.login(allow_tty_prompt=True)
        self.assertTrue(mock1.called)
        self.assertTrue(mock2.called)

    @mock.patch(
        "planet_auth.oidc.api_clients.token_api_client.TokenApiClient.get_token_from_password",
        mocked_tokenapi_password,
    )
    @mock.patch("getpass.getpass", side_effect=mocked_getpass_password)
    @mock.patch("builtins.input", side_effect=mocked_input_username)
    def test_login_explicit_username_password(self, mock1, mock2):
        self.under_test_without_username_password.login(
            username="test_username_arg", password="test_password_arg", allow_tty_prompt=True
        )
        self.assertFalse(mock1.called)
        self.assertFalse(mock2.called)

    def test_login_failure_username_input_needed(self):
        with self.assertRaises(ResourceOwnerAuthClientException):
            self.under_test_without_username_password.login(
                username=None, password="test_password", allow_tty_prompt=False
            )

    def test_login_failure_password_input_needed(self):
        with self.assertRaises(ResourceOwnerAuthClientException):
            self.under_test_without_username_password.login(
                username="test_user", password=None, allow_tty_prompt=False
            )

    def test_auth_enricher(self):
        # The correctness of what enrichment does is determined by the OAuth
        # endpoint, and we've externalized the heavy lifting of preparing the
        # enrichment into helper functions that could be unit tested
        # separately. (So unit testing of that ought to be done there,
        # although perhaps testing it here in the context of a particular
        # flow here is more meaningful.)
        enriched_payload, auth = self.under_test_with_username_password._client_auth_enricher({}, "test_audience")

        # Payload enriched with *only* the client id.
        self.assertEqual({"client_id": TEST_CLIENT_ID}, enriched_payload)

        # No request auth expected
        self.assertIsNone(auth)

    def test_default_request_authenticator_type(self):
        test_result = self.under_test_with_username_password.default_request_authenticator(
            credential=TEST_TOKEN_SAVE_FILE_PATH
        )
        self.assertIsInstance(test_result, RefreshOrReloginOidcTokenRequestAuthenticator)

    def test_default_request_authenticator_given_literal_credential_object_instead_of_path(self):
        literal_credential = STUB_AUTH_CLIENT.login()
        test_result = self.under_test_with_username_password.default_request_authenticator(
            credential=literal_credential
        )
        test_result.pre_request_hook()
        self.assertEqual(literal_credential.access_token(), test_result._token_body)

    def test_default_request_authenticator_given_invalid_type(self):
        literal_credential = FileBackedApiKey()
        with self.assertRaises(TypeError):
            self.under_test_with_username_password.default_request_authenticator(credential=literal_credential)


class ResourceOwnerWithClientSecretClientConfigTest(unittest.TestCase):
    def test_secret_required(self):
        # No exception
        under_test = ResourceOwnerWithClientSecretClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_secret=TEST_CLIENT_SECRET,
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
        )
        under_test.check()

        under_test = ResourceOwnerWithClientSecretClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


class ResourceOwnerWithClientSecretFlowTest(unittest.TestCase):
    def setUp(self):
        self.under_test = ResourceOwnerWithClientSecretAuthClient(
            ResourceOwnerWithClientSecretClientConfig(
                auth_server=TEST_AUTH_SERVER,
                client_id=TEST_CLIENT_ID,
                client_secret=TEST_CLIENT_SECRET,
                username=TEST_USER_NAME,
                password=TEST_USER_PASSWORD,
            )
        )

    def test_auth_enricher(self):
        # The correctness of what enrichment does is determined by the OAuth
        # endpoint, and we've externalized the heavy lifting of preparing the
        # enrichment into helper functions that could be unit tested
        # separately. (So unit testing of that ought to be done there,
        # although perhaps testing it here in the context of a particular
        # flow here is more meaningful.)
        enriched_payload, auth = self.under_test._client_auth_enricher({}, "test_audience")

        # No payload enrichment expected.
        self.assertEqual({}, enriched_payload)

        # HTTP basic auth with the client secret.
        self.assertEqual(TEST_CLIENT_ID, auth.username)
        self.assertEqual(TEST_CLIENT_SECRET, auth.password)


class ResourceOwnerWithPubKeyClientConfigTest(unittest.TestCase):
    def test_privkey_required(self):
        # No exception
        under_test = ResourceOwnerWithPubKeyClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_file="/dummy/utest/file",
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
        )
        under_test.check()

        # No exception
        under_test = ResourceOwnerWithPubKeyClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey="dummy private key literal",
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
        )
        under_test.check()

        under_test = ResourceOwnerWithPubKeyClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


class ResourceOwnerWithPubKeyFlowTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.privkey_password = "password"
        cls.privkey_file_path = tdata_resource_file_path("keys/keypair1_priv.test_pem")

    def setUp(self):
        self.under_test = ResourceOwnerWithPubKeyAuthClient(
            ResourceOwnerWithPubKeyClientConfig(
                auth_server=TEST_AUTH_SERVER,
                client_id=TEST_CLIENT_ID,
                client_privkey_password=self.privkey_password,
                client_privkey_file=self.privkey_file_path,
                username=TEST_USER_NAME,
                password=TEST_USER_PASSWORD,
            )
        )

    def test_auth_enricher(self):
        # The correctness of what enrichment does is determined by the OAuth
        # endpoint, and we've externalized the heavy lifting of preparing the
        # enrichment into helper functions that could be unit tested
        # separately. (So unit testing of that ought to be done there,
        # although perhaps testing it here in the context of a particular
        # flow here is more meaningful.)
        enriched_payload, auth = self.under_test._client_auth_enricher({}, "test_audience")

        # Payload enriched with a signed key assertion
        self.assertEqual(
            "urn:ietf:params:oauth:client-assertion-type:jwt-bearer", enriched_payload.get("client_assertion_type")
        )
        self.assertIsNotNone(enriched_payload.get("client_assertion"))

        # No request auth expected
        self.assertIsNone(auth)



import os
import sys

import nox

nox.options.stop_on_first_error = True
nox.options.reuse_existing_virtualenvs = False

# Default sessions - all tests, but not packaging
nox.options.sessions = [
    "black_lint",
    "pytest",
    "semgrep_src",
    "mypy",
    "pyflakes_src",
    "pyflakes_examples",
    "pyflakes_tests",
    "pylint_src",
    "pylint_examples",
    "pylint_tests",
]

_DEFAULT_PYTHON = "3.13"
_ALL_PYTHON = ["3.9", "3.10", "3.11", "3.12", "3.13"]


@nox.session(python=_ALL_PYTHON)
def pytest(session):
    """Run Pytest test suites"""
    session.install("-e", ".[test]")

    options = session.posargs
    if "-k" in options:
        options.append("--no-cov")
    # Default test set selection done in pyproject.toml
    # session.run("pytest", "--log-cli-level=DEBUG", "-v", *options)
    session.run("pytest", "-v", *options)


@nox.session(python=_DEFAULT_PYTHON)
def semgrep_src(session):
    """Scan the code for security problems with semgrep"""
    session.install("-e", ".[test]")
    # session.run("semgrep", "scan", "--strict", "--verbose", "--error", "--junit-xml", "--junit-xml-output=semgrep-src.xml", "src")
    session.run("semgrep", "scan", "--strict", "--verbose", "--error", "src")


@nox.session(python=_DEFAULT_PYTHON)
def black_lint(session):
    """Check code formatting with Black"""
    session.install("-e", ".[test]")
    session.run("black", "--verbose", "--check", "--diff", "--color", ".")


@nox.session(python=_DEFAULT_PYTHON)
def black_format(session):
    """Fix code formatting with Black"""
    session.install("-e", ".[test]")
    session.run("black", "--verbose", ".")


@nox.session(python=_DEFAULT_PYTHON)
def mypy(session):
    """Lint all of the code with mypy"""
    session.install("-e", ".[test, examples]")
    session.run("mypy", "--install-type", "--non-interactive", "--junit-xml", "mypy.xml")


@nox.session(python=_DEFAULT_PYTHON)
def pyflakes_src(session):
    """Lint the library code with Pyflakes"""
    session.install("-e", ".[test]")
    session.run("pyflakes", "src")


@nox.session(python=_DEFAULT_PYTHON)
def pyflakes_examples(session):
    """Lint the example code with Pyflakes"""
    session.install("-e", ".[test, examples]")
    session.run("pyflakes", "docs/examples")


@nox.session(python=_DEFAULT_PYTHON)
def pyflakes_tests(session):
    """Lint the test code with Pyflakes"""
    session.install("-e", ".[test]")
    session.run("pyflakes", "tests")


@nox.session(python=_DEFAULT_PYTHON)
def pylint_src(session):
    """Lint the library code with Pylint"""
    session.install("-e", ".[test]")
    session.run("pylint", "src")


@nox.session(python=_DEFAULT_PYTHON)
def pylint_examples(session):
    """Lint the example code with Pylint"""
    session.install("-e", ".[test, examples]")
    session.run("pylint", "docs/examples")


@nox.session(python=_DEFAULT_PYTHON)
def pylint_tests(session):
    """Lint the test code with Pylint"""
    session.install("-e", ".[test]")
    session.run("pylint", "--disable", "protected-access", "--disable", "unused-variable", "tests")


@nox.session(python=_DEFAULT_PYTHON)
def pkg_build_wheel(session):
    """Build distribution package files"""
    session.install("-e", ".[build]")
    session.run("pyproject-build")
    # session.run("simple503", "-B", "dist", "dist")


@nox.session(python=_DEFAULT_PYTHON)
def pkg_build_local_dist(session):
    """Build distribution package files, and build a local simple PyPi directory that can be used by pip for local testing."""
    session.install("-e", ".[build]")
    session.run("pyproject-build")
    session.run("simple503", "-B", "dist", "dist")


@nox.session(python=_DEFAULT_PYTHON)
def pkg_check(session):
    """Check the built distribution files for errors"""
    session.install("-e", ".[build]")
    session.run("twine", "check", "--strict", "dist/*.whl", "dist/*.tar.gz")


def _publish_pypi(session, repo_url, token):
    session.install("-e", ".[build]")
    session.run(
        "twine",
        "upload",
        "--non-interactive",
        # "--verbose",
        "--username",
        "__token__",
        "--password",
        token,
        "--repository-url",
        repo_url,
        # "--repository",
        # repo_name,
        "dist/*.whl",
        "dist/*.tar.gz",
    )


@nox.session(python=_DEFAULT_PYTHON)
def pkg_publish_pypi_prod(session):
    """Publish packages to the production PyPi server"""
    token = os.getenv("NOX_PYPI_API_TOKEN")
    if not token:
        sys.exit("NOX_PYPI_API_TOKEN must be set in the environment with the PyPi access token")
    _publish_pypi(session, "https://upload.pypi.org/legacy", token)


@nox.session(python=_DEFAULT_PYTHON)
def pkg_publish_pypi_test(session):
    """Publish packages to the test PyPi server"""
    token = os.getenv("NOX_PYPI_API_TOKEN")
    if not token:
        sys.exit("NOX_PYPI_API_TOKEN must be set in the environment with the PyPi access token")
    _publish_pypi(session, "https://test.pypi.org/legacy/", token)


@nox.session(python=_DEFAULT_PYTHON)
def mkdocs_build(session):
    """Build the documentation locally"""
    session.install("-e", ".[docs]")
    session.run("mkdocs", "-v", "build", "--clean")


@nox.session(python=_DEFAULT_PYTHON)
def mkdocs_serve(session):
    """Build the documentation and serve locally over HTTP. The server will watch for updates."""
    session.install("-e", ".[docs]")
    session.run("mkdocs", "-v", "serve")


@nox.session(python=_DEFAULT_PYTHON)
def mkdocs_publish_readthedocs(session):
    """(NOT IMPLEMENTED) Publish the documentation to ReadTheDocs.com"""
    session.install("-e", ".[build, docs]")
    # TODO - Manual doc publishing
    print(
        "ERROR: Read The Docs publishing not implemented in the noxfile."
        "  Documentation publishing is triggered via GitHub webhook."
    )
    assert False



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import unittest
from unittest import mock

from planet_auth import FileBackedApiKey
from planet_auth.auth_client import AuthClientConfigException
from planet_auth.oidc.auth_clients.client_credentials_flow import (
    ClientCredentialsPubKeyAuthClient,
    ClientCredentialsPubKeyClientConfig,
    ClientCredentialsClientSecretAuthClient,
    ClientCredentialsClientSecretClientConfig,
)
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.request_authenticator import RefreshOrReloginOidcTokenRequestAuthenticator
from tests.test_planet_auth.unit.auth.util import StubOidcClientConfig, StubOidcAuthClient
from tests.test_planet_auth.util import tdata_resource_file_path

TEST_AUTH_SERVER = "https://blackhole.unittest.planet.com/fake_authserver"
TEST_CLIENT_ID = "fake_test_client_id"
TEST_CLIENT_SECRET = "fake_test_client_secret"
TEST_TOKEN_SAVE_FILE_PATH = pathlib.Path("/test/token.json")
MOCK_TOKEN = {"token_type": "Bearer", "expires_in": 3600, "access_token": "__mock_access_token__", "scope": "planet"}

# Get some tokens that look more real than the mock token for some tests.
TEST_SIGNING_KEY_FILE = tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem")
TEST_SIGNING_PUBKEY_FILE = tdata_resource_file_path("keys/keypair1_pub_jwk.json")
STUB_AUTH_CLIENT_CONFIG = StubOidcClientConfig(
    auth_server=TEST_AUTH_SERVER,
    stub_authority_ttl=3600,
    stub_authority_access_token_audience="utest_audience",
    stub_authority_signing_key_file=TEST_SIGNING_KEY_FILE,
    stub_authority_pub_key_file=TEST_SIGNING_PUBKEY_FILE,
    scopes=["offline_access", "profile", "openid", "test_scope_1", "test_scope_2"],
)
STUB_AUTH_CLIENT = StubOidcAuthClient(STUB_AUTH_CLIENT_CONFIG)


def mocked_tokenapi_ccred(obj_self, client_id, requested_scopes, requested_audiences, auth_enricher, extra):
    return MOCK_TOKEN


class ClientCredentialsClientSecretConfigTest(unittest.TestCase):
    def test_secret_required(self):
        # No exception
        under_test = ClientCredentialsClientSecretClientConfig(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, client_secret=TEST_CLIENT_SECRET
        )
        under_test.check()

        under_test = ClientCredentialsClientSecretClientConfig(auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID)
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


class ClientCredentialsClientSecretFlowTest(unittest.TestCase):
    def setUp(self):
        self.under_test = ClientCredentialsClientSecretAuthClient(
            ClientCredentialsClientSecretClientConfig(
                auth_server=TEST_AUTH_SERVER,
                token_endpoint=TEST_AUTH_SERVER + "/token",
                client_id=TEST_CLIENT_ID,
                client_secret=TEST_CLIENT_SECRET,
            )
        )

    @mock.patch(
        "planet_auth.oidc.api_clients.token_api_client.TokenApiClient.get_token_from_client_credentials",
        mocked_tokenapi_ccred,
    )
    def test_login(self):
        test_result = self.under_test.login()
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        self.assertEqual(MOCK_TOKEN, test_result.data())

        # again with override scopes, but since the response is mocked
        # there is nothing different to check in the result data.
        # Same goes for audiences.
        test_result = self.under_test.login(requested_scopes=["override1"], requested_audiences=["req_aud1"])
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        # self.assertEqual(MOCK_TOKEN, test_result.data())

    def test_default_request_authenticator_type(self):
        test_result = self.under_test.default_request_authenticator(credential=TEST_TOKEN_SAVE_FILE_PATH)
        self.assertIsInstance(test_result, RefreshOrReloginOidcTokenRequestAuthenticator)

    def test_auth_enricher(self):
        # The correctness of what enrichment does is determined by the OAuth
        # endpoint, and we've externalized the heavy lifting of preparing the
        # enrichment into helper functions that could be unit tested
        # separately. (So unit testing of that ought to be done there,
        # although perhaps testing it here in the context of a particular
        # flow here is more meaningful.)
        enriched_payload, auth = self.under_test._client_auth_enricher({}, "test_audience")

        # No payload enrichment expected.
        self.assertEqual({}, enriched_payload)

        # HTTP basic auth with the client secret.
        self.assertEqual(TEST_CLIENT_ID, auth.username)
        self.assertEqual(TEST_CLIENT_SECRET, auth.password)

    def test_auth_login_enricher(self):
        # Client secret client credentials uses a different enricher
        # for logins vs more other use cases.  Again, it's not called
        # by the class directly, but by an API client that is mocked out
        # for unit tests. Again, the arbiter of what is correct is
        # the external token server.
        enriched_payload, enriched_auth = self.under_test._client_auth_enricher_login({}, "test_audience")

        # Client secret ought to be put into the payload
        self.assertEqual({"client_id": TEST_CLIENT_ID, "client_secret": TEST_CLIENT_SECRET}, enriched_payload)
        self.assertIsNone(enriched_auth)


class ClientCredentialsPubKeyClientConfigTest(unittest.TestCase):
    def test_privkey_required(self):
        # No exception
        under_test = ClientCredentialsPubKeyClientConfig(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, client_privkey_file="/dummy/utest/file"
        )
        under_test.check()

        # No exception
        under_test = ClientCredentialsPubKeyClientConfig(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, client_privkey="dummy private key literal"
        )
        under_test.check()

        under_test = ClientCredentialsPubKeyClientConfig(auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID)
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


class ClientCredentialsPubKeyFlowTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.privkey_password = "password"
        cls.privkey_file_path = tdata_resource_file_path("keys/keypair1_priv.test_pem")

    def setUp(self):
        self.under_test = ClientCredentialsPubKeyAuthClient(
            ClientCredentialsPubKeyClientConfig(
                auth_server=TEST_AUTH_SERVER,
                token_endpoint=TEST_AUTH_SERVER + "/token",
                client_id=TEST_CLIENT_ID,
                client_privkey_password=self.privkey_password,
                client_privkey_file=self.privkey_file_path,
            )
        )

    @mock.patch(
        "planet_auth.oidc.api_clients.token_api_client.TokenApiClient.get_token_from_client_credentials",
        mocked_tokenapi_ccred,
    )
    def test_login(self):
        test_result = self.under_test.login()
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        self.assertEqual(MOCK_TOKEN, test_result.data())

        # again with override scopes, but since the response is mocked
        # there is nothing different to check in the result data/
        # Same goes for audiences.
        test_result = self.under_test.login(requested_scopes=["override1"], requested_audiences=["req_aud1"])
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        # self.assertEqual(MOCK_TOKEN, test_result.data())

    def test_default_request_authenticator_type(self):
        test_result = self.under_test.default_request_authenticator(credential=TEST_TOKEN_SAVE_FILE_PATH)
        self.assertIsInstance(test_result, RefreshOrReloginOidcTokenRequestAuthenticator)

    def test_default_request_authenticator_given_literal_credential_object_instead_of_path(self):
        literal_credential = STUB_AUTH_CLIENT.login()
        test_result = self.under_test.default_request_authenticator(credential=literal_credential)
        test_result.pre_request_hook()
        self.assertEqual(literal_credential.access_token(), test_result._token_body)

    def test_default_request_authenticator_given_invalid_type(self):
        literal_credential = FileBackedApiKey()
        with self.assertRaises(TypeError):
            self.under_test.default_request_authenticator(credential=literal_credential)

    def test_auth_enricher(self):
        # The correctness of what enrichment does is determined by the OAuth
        # endpoint, and we've externalized the heavy lifting of preparing the
        # enrichment into helper functions that could be unit tested
        # separately. (So unit testing of that ought to be done there,
        # although perhaps testing it here in the context of a particular
        # flow here is more meaningful.)
        enriched_payload, auth = self.under_test._client_auth_enricher({}, "test_audience")

        # Payload enriched with a signed key assertion
        self.assertEqual(
            "urn:ietf:params:oauth:client-assertion-type:jwt-bearer", enriched_payload.get("client_assertion_type")
        )
        self.assertIsNotNone(enriched_payload.get("client_assertion"))

        # No request auth expected
        self.assertIsNone(auth)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import hashlib
import secrets

_DEFAULT_LENGTH = 96


def generate_nonce(length):
    return "".join([str(secrets.randbelow(10)) for i in range(length)])


def create_challenge(v):
    code_challenge = hashlib.sha256(v.encode("utf-8")).digest()
    code_challenge = base64.urlsafe_b64encode(code_challenge).decode("utf-8")
    code_challenge = code_challenge.replace("=", "")
    return code_challenge


def create_verifier(length=_DEFAULT_LENGTH):
    code_verifier = secrets.token_urlsafe(length)
    return code_verifier


def create_pkce_challenge_verifier_pair(length=_DEFAULT_LENGTH):
    v = create_verifier(length)
    c = create_challenge(v)
    return v, c



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from click.testing import CliRunner

from planet_auth_utils.commands.cli.main import cmd_plauth


@pytest.fixture(name="click_cli_runner")
def fixture_click_cli_runner():
    return CliRunner()


class TestPlauthCli:
    # Smoke tests are to insure the commands are not just exception throwing
    # and DOA when invoked.  We do not currently test all commands like this,
    # but we do check the bare minimum of making sure each command group
    # works (which should show help).  This is to insure we do not break things
    # when refactoring or updating dependencies at a distance.

    @pytest.fixture(autouse=True)
    def _setup_caplog_fixture(self, caplog):
        # Click and Pytest seem to have some bad interactions with logs and
        # unit testing that lead to "ValueError: I/O operation on closed file"
        # when commands invoke logging.  Effectively disable logging by
        # setting the level insanely high.
        self._caplog = caplog  # pylint: disable=W0201
        self._caplog.set_level(100000)

    def test_main_no_option_shows_help(self, click_cli_runner):
        result = click_cli_runner.invoke(cmd_plauth)
        assert 0 == result.exit_code
        assert "Planet authentication utility" in result.stdout

    def test_smoke_legacy(self, click_cli_runner):
        result = click_cli_runner.invoke(cmd_plauth, ["legacy"])
        assert 0 == result.exit_code

    def test_smoke_oauth(self, click_cli_runner):
        result = click_cli_runner.invoke(cmd_plauth, ["oauth"])
        assert 0 == result.exit_code

    def test_smoke_profile(self, click_cli_runner):
        result = click_cli_runner.invoke(cmd_plauth, ["profile"])
        assert 0 == result.exit_code

    def test_smoke_version(self, click_cli_runner):
        result = click_cli_runner.invoke(cmd_plauth, ["version"])
        assert 0 == result.exit_code



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jwt
import time
from typing import Dict

import planet_auth.logging.auth_logger
from planet_auth.credential import Credential
from planet_auth.request_authenticator import CredentialRequestAuthenticator
from planet_auth.oidc.auth_client import OidcAuthClient
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


# Note: Auth client can be passed override scopes for token refresh, but we
#       haven't plumbed our auth helpers to do this. This is probably an
#       acceptable limitation for this use case.
# Note: We assume the caller knows what they are doing. It might be
#       nice to warn (or fail) if we detect that there is a mismatch between
#       the destination this authenticator is providing tokens to and the
#       audience of the token.  That is beyond basic OAuth, since
#       1) tokens need not be transparent or JWTs, and
#       2) the audience need not be a URL we understand.
#       The risk is in handing out access tokens to a destination
#       that we do not trust.
class RefreshingOidcTokenRequestAuthenticator(CredentialRequestAuthenticator):
    """
    Decorate a http request with an OAuth bearer auth token. Automatically
    initiate a refresh request if we know the access token to be close to
    expiration.

    This class assumes access tokens are JWTs and can be locally inspected,
    which OIDC and OAuth do not require.  JWT access tokens that comply with
    RFC 9068 can be checked for expiration timing without hitting the network
    token introspection endpoint.
    """

    def __init__(self, credential: FileBackedOidcCredential, auth_client: OidcAuthClient = None, **kwargs):
        super().__init__(credential=credential, **kwargs)
        self._auth_client = auth_client
        self._refresh_at = 0

    def _load(self):
        if self._credential.path():
            # allow in memory operation.
            self._credential.load()

        access_token_str = self._credential.access_token()
        # Absolutely not appropriate to not verify the signature in a token
        # validation context (e.g. server side auth of a client). Here we
        # know that's not what we are doing. This is a client helper class
        # for clients who will be presenting tokens to such a server.  We
        # are inspecting ourselves, not verifying for trust purposes.
        # We are not expected to be the audience.
        # TODO: we should use `expires_in` from the response from the
        #    OAuth server, which would work for non-JWT opaque OAuth
        #    tokens.  Since that is a relative time, we would also need
        #    to augment our FileBackedOidcCredential with an issued
        #    at time.
        unverified_decoded_atoken = jwt.decode(access_token_str, options={"verify_signature": False})  # nosemgrep
        iat = unverified_decoded_atoken.get("iat") or 0
        exp = unverified_decoded_atoken.get("exp") or 0
        # refresh at the 3/4 life
        self._refresh_at = int(iat + (3 * (exp - iat) / 4))
        self._token_body = access_token_str

    def _refresh(self):
        if self._auth_client:
            new_credentials = self._auth_client.refresh(self._credential.refresh_token())
            new_credentials.set_path(self._credential.path())
            new_credentials.set_storage_provider(self._credential.storage_provider())
            new_credentials.save()

            self._credential = new_credentials
            self._load()

    def pre_request_hook(self):
        # Reload the file before refreshing. Another process might
        # have done it for us, and save us the network call.
        #
        # Also, if refresh tokens are configured to be one time use,
        # we want a fresh refresh token. Stale refresh tokens are
        # invalid in this case.
        #
        # Also, it's possible that we have a valid refresh token,
        # but not an access token.  When that's true, we should
        # try to cash in the refresh token.
        #
        # If everything fails, continue with what we have. Let the API
        # we are calling decide if it's good enough.
        if int(time.time()) > self._refresh_at:
            try:
                self._load()
            except Exception as e:  # pylint: disable=broad-exception-caught
                auth_logger.warning(
                    msg="Error loading auth token. Continuing with old auth token. Load error: " + str(e)
                )
        if int(time.time()) > self._refresh_at:
            try:
                self._refresh()
            except Exception as e:  # pylint: disable=broad-exception-caught
                auth_logger.warning(
                    msg="Error refreshing auth token. Continuing with old auth token. Refresh error: " + str(e)
                )
        super().pre_request_hook()

    def update_credential(self, new_credential: Credential):
        if not isinstance(new_credential, FileBackedOidcCredential):
            raise TypeError(
                f"{type(self).__name__} does not support {type(new_credential)} credentials.  Use FileBackedOidcCredential."
            )
        super().update_credential(new_credential=new_credential)
        self._refresh_at = 0
        # self._load()  # Mimic __init__.  Don't load, let that happen JIT.

    def update_credential_data(self, new_credential_data: Dict):
        # This is more different than update_credential() than it may
        # appear. Inherent in being passed a Credential in update_credential()
        # is that it may not yet be loaded from disk, and so deferring
        # the _load() as a JIT operation is appropriate.  In this case,
        # being passed the data struct is as if a network refresh call
        # has already taken place or a _load() is in progress, and we
        # should behave as  if we are finishing a _refresh() call.
        super().update_credential_data(new_credential_data=new_credential_data)
        self._load()


class RefreshOrReloginOidcTokenRequestAuthenticator(RefreshingOidcTokenRequestAuthenticator):
    """
    Decorate a http request with a bearer auth token. Automatically initiate
    a refresh request using the refresh token if we know the access token to
    be close to expiration. If we do not have a refresh token, then fall back
    to re-initiating a login.  Sometimes (like with client credential flow),
    refresh tokens may not be available and we might want to login rather
    than refresh.  It is not an automatic choice to fall back to login when
    refresh is not available, since for some auth client configurations login
    is interactive, and would not be appropriate for headless use cases.
    Refresh should always be silent.
    """

    def __init__(self, credential: FileBackedOidcCredential, auth_client: OidcAuthClient = None, **kwargs):
        super().__init__(credential=credential, auth_client=auth_client, **kwargs)

    def _refresh(self):
        if self._auth_client:
            try:
                _refresh_token = self._credential.refresh_token()
            except FileNotFoundError:
                # Token file might not exist in the "or (re)login" case.
                _refresh_token = None

            if _refresh_token:
                new_credentials = self._auth_client.refresh(_refresh_token)
            else:
                new_credentials = self._auth_client.login(allow_open_browser=False, allow_tty_prompt=False)

            new_credentials.set_path(self._credential.path())
            new_credentials.set_storage_provider(self._credential.storage_provider())
            new_credentials.save()

            self._credential = new_credentials
            self._load()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod, ABC
from typing import Dict

import httpx
import requests.auth

import planet_auth.logging.auth_logger
from planet_auth.auth_exception import AuthException
from planet_auth.constants import X_PLANET_APP_HEADER, X_PLANET_APP
from planet_auth.credential import Credential

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class RequestAuthenticator(requests.auth.AuthBase, httpx.Auth, ABC):
    """
    Decorate a http request with a bearer auth token.
    """

    def __init__(self, token_body: str = None, token_prefix: str = "Bearer", auth_header: str = "Authorization"):
        # requests.auth.AuthBase.__init__(self)  # Known to be a no-op
        # httpx.Auth.__init__(self)  # Known to be a no-op
        self._token_prefix = token_prefix
        self._token_body = token_body
        self._auth_header = auth_header

    @abstractmethod
    def pre_request_hook(self):
        """
        Hook that will be called immediately prior to making an HTTP request
        so that implementing classes may make preparations.  Derived
        classes are expected to populate the member fields _token_prefix,
        _token_body, and _auth_header with values that are appropriate to
        the specific implementation.  These will then be used during
        subsequent HTTP request to authenticate the connection using
        a beater token authorization HTTP header.

        Implementers may make external network calls as required to perform
        necessary tasks such as refreshing access tokens.

        Implementations should not require user interaction by default. If
        an authentication mechanism will require user interaction, this
        should be an explicit decision that is left to the application
        using the RequestAuthenticator to control.
        """

    def _build_auth_header_payload(self):
        if self._token_prefix:
            # Should we make the space part of the prefix?  What if someone
            # wants no space?
            return self._token_prefix + " " + self._token_body
        else:
            return self._token_body

    def __call__(self, r):
        """
        Decorate a "requests" library based request with authentication
        """
        self.pre_request_hook()
        if self._token_body:
            r.headers[self._auth_header] = self._build_auth_header_payload()
        if X_PLANET_APP_HEADER not in r.headers:
            r.headers[X_PLANET_APP_HEADER] = X_PLANET_APP
        return r

    def auth_flow(self, request: httpx._models.Request):
        """
        Decorate a "httpx" library based request with authentication
        """
        self.pre_request_hook()
        if self._token_body:
            request.headers[self._auth_header] = self._build_auth_header_payload()
        if X_PLANET_APP_HEADER not in request.headers:
            request.headers[X_PLANET_APP_HEADER] = X_PLANET_APP
        yield request


class CredentialRequestAuthenticator(RequestAuthenticator, ABC):
    def __init__(self, credential: Credential = None, **kwargs):
        super().__init__(**kwargs)
        self._credential = credential

    def update_credential(self, new_credential: Credential):
        """
        Update the request authenticator with a new credential.
        It is the job of a derived class to know how to map the contents
        of a credential into the primitives known to HTTP request
        auth mechanics.  Not all derived classes are expected to
        understand all credential types.
        """
        self._credential = new_credential
        # Clear-out auth material when a new credential is set.
        # child classes are expected to populate it JIT for auth
        # requests.
        self._token_body = None

    def update_credential_data(self, new_credential_data: Dict):
        """
        Provide raw data that should be used to update the Credential
        object used to authenticate requests.  This information will
        be treated as newer than any file back data associated
        with the credential held by the authenticator, and the file
        will be overwritten.
        """
        if not self._credential:
            # The use cases for this being unset are narrow - mostly in the noop_auth.py module,
            # and the service side client_validator.py.  Neither of these use cases should
            # ever need to call update_credential_data().
            # Should we fail loud, or quiet?  Loud for now.
            raise RuntimeError("Programming error. Cannot update a credential that has has never been set.")

        self._credential.set_data(new_credential_data)
        self._credential.save()  # Clobber old data that may be saved to disk.
        # Clear-out auth material when a new credential is set.
        # child classes are expected to populate it JIT for auth
        # requests.
        self._token_body = None

    def credential(self):
        """
        Return the current credential.

        This may not be the credential the authenticator was constructed with.
        Request Authenticators are free to refresh credentials depending in the
        needs of the implementation.
        """
        return self._credential

    def is_initialized(self) -> bool:
        """
        Return true if the CredentialRequestAuthenticator has a credential that
        can be used to make requests.  If a credential is not in memory, storage
        is checked for the existence of a credential, but it will not be loaded
        from storage at this time, preserving JIT loading behavior.
        """
        if self._credential:
            return bool(self._credential.data()) or self._credential.is_persisted_to_storage()
        return False


class SimpleInMemoryRequestAuthenticator(CredentialRequestAuthenticator):
    # This SimpleInMemoryRequestAuthenticator subclasses from
    # CredentialRequestAuthenticator so it can fit in places that
    # is needed. It does not actually know about any Credential
    # types, since how credential types map to the HTTP request
    # authentication fields is credential specific.  This class
    # is more useful for testing and stubbing out interfaces.
    def pre_request_hook(self):
        pass

    def update_credential(self, new_credential: Credential):
        auth_logger.warning(msg="Generic SimpleInMemoryRequestAuthenticator ignores update_credential()")


class ForbiddenRequestAuthenticator(CredentialRequestAuthenticator):
    def pre_request_hook(self):
        raise AuthException(
            message="Making authenticated requests with the ForbiddenRequestAuthenticator is forbidden."
            " This is most likely the result of a configuration or programming error."
        )



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import importlib.resources

from planet_auth.oidc import resources as planet_auth_oidc_resources


class TestResources(unittest.TestCase):
    def test_can_read_oidc_resources(self):
        # I've broken resource packaging so many times....
        # This is best run against an installed dist package.
        # Running against the local dev tree doesn't catch
        # all problems (like a missed MANIFEST.in file)
        under_test = (
            importlib.resources.files(planet_auth_oidc_resources)
            .joinpath("callback_acknowledgement.html")
            .read_text("utf-8")
        )
        self.assertIsNotNone(under_test)
